from typing import Any

import aiomqtt
import structlog

from ...core.zmq.message import EventMessage  # noqa: F401
from ...serializers import Serializers
from ...td import EventAffordance, PropertyAffordance
from ..repository import BrokerThing  # noqa: F401


class TopicPublisher:
    """
    Publishes an event to an MQTT topic. Supply a different class in `MQTTPublisher` to use a different one.
    This object would be a controller in layered architecture.
    """

    def __init__(
        self,
        client: aiomqtt.Client,
        resource: EventAffordance | PropertyAffordance,
        config: Any,
        logger: structlog.stdlib.BoundLogger,
    ) -> None:
        """
        Parameters
        ----------
        client: aiomqtt.Client
            The MQTT client to use for publishing messages
        resource: EventAffordance | PropertyAffordance
            dataclass representation of observable property or event to be published
        config: RuntimeConfig
            The runtime configuration for the `MQTTPublisher`
        logger: structlog.stdlib.BoundLogger
            The logger to use for logging messages
        """
        from .config import RuntimeConfig  # noqa: F401

        self.client = client
        self.resource = resource
        self.topic = f"{self.resource.thing_id}/{self.resource.name}"
        self.config = config  # type: RuntimeConfig
        self.logger = logger.bind(layer="controller", impl=self.__class__.__name__, topic=self.topic)
        self.thing = self.config.thing_repository[resource.thing_id]  # type: BrokerThing
        self.qos = self.config.qos
        self._stop_publishing = False

    def stop(self):
        """stop publishing, the client is not closed automatically"""
        self._stop_publishing = True

    async def publish(self):
        """Publishes events to the MQTT broker in an infinite loop"""
        consumer = self.thing.subscribe_event(self.resource)
        self.logger.info(f"Starting to publish events for {self.resource.name} to MQTT broker on topic {self.topic}")
        while not self._stop_publishing:
            try:
                message = await consumer.receive()  # type: EventMessage | None
                if message is None:
                    continue
                payload = self.thing.get_response_payload(message)
                await self.client.publish(
                    topic=self.topic,
                    payload=payload.value,
                    qos=self.qos,
                    properties=dict(content_type=payload.content_type),
                )
                self.logger.debug(f"Published MQTT message for {self.resource.name} on topic {self.topic}")
            except Exception as ex:
                self.logger.error(f"Error publishing MQTT message for {self.resource.name}: {ex}")
        self.logger.info(f"Stopped publishing events for {self.resource.name} to MQTT broker on topic {self.topic}")


class ThingDescriptionPublisher:
    """
    Publishes Thing Description to an MQTT Topic. Supply a different class in `MQTTPublisher` to use a different one.
    This object would be a controller in layered architecture.
    """

    def __init__(
        self,
        client: aiomqtt.Client,
        config: Any,
        logger: structlog.stdlib.BoundLogger,
        ZMQ_TD: dict[str, Any],
    ) -> None:
        """
        Parameters
        ----------
        client: aiomqtt.Client
            The MQTT client to use for publishing messages
        config: RuntimeConfig
            The runtime configuration for the MQTT publisher
        logger: structlog.stdlib.BoundLogger
            The logger to use for logging messages
        ZMQ_TD: dict[str, Any]
            The ZMQ Thing Description message received from ZMQ broker
        """
        from .config import RuntimeConfig  # noqa: F401

        self.client = client
        self.topic = f"{ZMQ_TD['id']}/thing-description"
        self.config = config  # type: RuntimeConfig
        self.logger = logger.bind(layer="controller", impl=self.__class__.__name__)
        self.thing = self.config.thing_repository[ZMQ_TD["id"]]
        self.thing_description = self.config.thing_description_service(
            hostname=self.client._hostname,
            port=self.client._port,
            logger=logger,
            ssl=self.client._client._ssl_context is not None,
        )

    async def publish(self, ZMQ_TD: dict[str, Any]) -> dict[str, Any]:
        """Publishes Thing Description to the MQTT broker, one-time at startup, with qos=2 and retain=True"""
        TD = await self.thing_description.generate(ZMQ_TD)

        await self.client.publish(
            topic=self.topic,
            payload=Serializers.json.dumps(TD),
            qos=2,
            properties=dict(content_type="application/json"),
            retain=True,
        )

        self.logger.info(f"Published Thing Description for {TD['id']} to MQTT broker on topic {self.topic}")



from dataclasses import asdict, dataclass
from typing import Any

from sqlalchemy import JSON, Integer, LargeBinary, String
from sqlalchemy.orm import (
    DeclarativeBase,
    Mapped,
    MappedAsDataclass,
    mapped_column,
)

from ..constants import JSONSerializable


class ThingTableBase(DeclarativeBase):
    """SQLAlchemy base table for all Thing related tables"""

    pass


class SerializedProperty(MappedAsDataclass, ThingTableBase):
    """
    Property value is serialized before storing in database, therefore providing unified version for
    SQLite and other relational tables
    """

    __tablename__ = "properties"

    id: Mapped[int] = mapped_column(Integer, primary_key=True, autoincrement=True)
    name: Mapped[str] = mapped_column(String, unique=True, nullable=False)
    serialized_value: Mapped[bytes] = mapped_column(LargeBinary, nullable=False)
    thing_id: Mapped[str] = mapped_column(String)
    thing_class: Mapped[str] = mapped_column(String)
    created_at: Mapped[str] = mapped_column(String)
    updated_at: Mapped[str] = mapped_column(String)
    content_type: Mapped[str] = mapped_column(String, nullable=False, default="application/json")


class ThingInformation(MappedAsDataclass, ThingTableBase):
    """Stores information about the Thing instance itself, useful metadata which may be later populated in a GUI"""

    __tablename__ = "things"

    id: Mapped[int] = mapped_column(Integer, primary_key=True, autoincrement=True)
    thing_id: Mapped[str] = mapped_column(String, unique=True, nullable=False)
    thing_class: Mapped[str] = mapped_column(String, nullable=False)
    script: Mapped[str] = mapped_column(String)
    init_kwargs: Mapped[JSONSerializable] = mapped_column(JSON)
    server_id: Mapped[str] = mapped_column(String)

    def json(self):
        return asdict(self)


@dataclass
class DeserializedProperty:  # not part of database
    """Property with deserialized value after fetching from database"""

    thing_id: str
    name: str
    value: Any
    created_at: str
    updated_at: str



from typing import Any, Callable

import aiomqtt
import structlog

from paho.mqtt.client import Client as PahoMQTTClient
from paho.mqtt.client import MQTTMessage

from ...serializers import BaseSerializer, Serializers  # noqa: F401
from ...td.forms import Form
from ...td.interaction_affordance import EventAffordance, PropertyAffordance
from ..abstractions import SSE, ConsumedThingEvent


class MQTTConsumer(ConsumedThingEvent):
    # An MQTT event consumer, both sync and async,
    # please dont add classdoc

    def __init__(
        self,
        sync_client: PahoMQTTClient,
        async_client: aiomqtt.Client,
        resource: EventAffordance | PropertyAffordance,
        qos: int,
        logger: structlog.stdlib.BoundLogger,
        owner_inst: Any,
    ) -> None:
        """
        Parameters
        ----------
        sync_client: PahoMQTTClient
            synchronous MQTT client
        async_client: aiomqtt.Client
            asynchronous MQTT client
        resource: EventAffordance | PropertyAffordance
            the event affordance to consume
        qos: int
            The MQTT QoS level to use
        logger: structlog.stdlib.BoundLogger
            Logger instance
        owner_inst: Any
            The parent object that owns this consumer
        """
        super().__init__(resource=resource, logger=logger, owner_inst=owner_inst)
        self.qos = qos
        self.sync_client = sync_client
        self.async_client = async_client
        self.subscribed = True

    def listen(self, form: Form, callbacks: list[Callable], concurrent: bool, deserialize: bool) -> None:
        # This method is called from a different thread but also finishes quickly, we wont redo this way
        # for the time being.
        topic = form.mqv_topic or f"{self.resource.thing_id}/{self.resource.name}"

        def on_topic_message(client: PahoMQTTClient, userdata, message: MQTTMessage):
            try:
                payload = message.payload
                # message.properties.readProperty("content_type") if message.properties else form.contentType
                # TODO, fix this, make sure to that content_type is not empty after extracting
                content_type = form.contentType or "application/json"
                serializer = Serializers.content_types.get(content_type, None)  # type: BaseSerializer
                if deserialize and content_type and serializer:
                    try:
                        payload = serializer.loads(payload)
                    except Exception as ex:
                        self.logger.error(
                            f"Error deserializing MQTT message for topic {topic}, "
                            + f"passing payload as it is. message: {ex}"
                        )
                        self.logger.exception(ex)
                event_data = SSE()
                event_data.data = payload
                event_data.id = message.mid
                self.schedule_callbacks(callbacks=callbacks, event_data=event_data, concurrent=concurrent)
            except Exception as ex:
                self.logger.error(f"Error handling MQTT message for topic {topic}: {ex}")
                self.logger.exception(ex)

        self.sync_client.message_callback_add(topic, on_topic_message)

    async def async_listen(self, form: Form, callbacks: list[Callable], concurrent: bool, deserialize: bool) -> None:
        topic = form.mqv_topic or f"{self.resource.thing_id}/{self.resource.name}"
        try:
            await self.async_client.__aenter__()
        except aiomqtt.MqttReentrantError:
            pass
        await self.async_client.subscribe(topic, qos=self.qos)
        async for message in self.async_client.messages:
            if not self.subscribed:
                break
            if not message.topic.matches(topic):
                continue
            try:
                payload = message.payload
                # message.properties.readProperty("content_type") if message.properties else form.contentType
                # TODO, fix this, make sure to that content_type is not empty after extracting
                content_type = form.contentType or "application/json"
                serializer = Serializers.content_types.get(content_type, None)  # type: BaseSerializer
                if deserialize and content_type and serializer:
                    try:
                        payload = serializer.loads(payload)
                    except Exception as ex:
                        self.logger.error(
                            f"Error deserializing MQTT message for topic {topic}, "
                            + f"passing payload as it is. message: {ex}"
                        )
                        self.logger.exception(ex)
                event_data = SSE()
                event_data.data = payload
                event_data.id = message.mid
                await self.async_schedule_callbacks(callbacks=callbacks, event_data=event_data, concurrent=concurrent)
            except Exception as ex:
                self.logger.error(f"Error handling MQTT message for topic {topic}: {ex}")
                self.logger.exception(ex)
        self.async_client.unsubscribe(topic)

    def unsubscribe(self) -> None:
        self.subscribed = False
        self.sync_client.message_callback_remove(f"{self.resource.thing_id}/{self.resource.name}")



import pytest
import zmq.asyncio

from hololinked.config import global_config
from hololinked.constants import ZMQ_TRANSPORTS
from hololinked.core.zmq.brokers import BaseZMQ


def test_01_socket_creation_defaults():
    """check the default settings of socket creation - an IPC socket which is a ROUTER and async"""
    socket, socket_address = BaseZMQ.get_socket(
        server_id="test-server",
        socket_id="test-server",
        node_type="server",
        context=global_config.zmq_context(),
    )
    assert isinstance(socket, zmq.asyncio.Socket)
    assert socket.getsockopt_string(zmq.IDENTITY) == "test-server"
    assert socket.socket_type == zmq.ROUTER
    assert socket_address.startswith("ipc://")
    assert socket_address.endswith(".ipc")
    socket.close()


def test_02_context_options():
    """
    Check that context and socket type are as expected.
    Async context should be used for async socket and sync context for sync socket.
    """
    context = zmq.Context()
    socket, _ = BaseZMQ.get_socket(
        server_id="test-server",
        socket_id="test-server",
        node_type="server",
        context=context,
    )
    assert isinstance(socket, zmq.Socket)
    assert not isinstance(socket, zmq.asyncio.Socket)
    socket.close()
    context.term()

    context = zmq.asyncio.Context()
    socket, _ = BaseZMQ.get_socket(
        server_id="test-server",
        socket_id="test-server",
        node_type="server",
        context=context,
    )
    assert isinstance(socket, zmq.Socket)
    assert isinstance(socket, zmq.asyncio.Socket)
    socket.close()
    context.term()


def test_03_transport_options():
    """check only three transport options are supported"""
    context = zmq.asyncio.Context()
    socket, socket_address = BaseZMQ.get_socket(
        server_id="test-server",
        socket_id="test-server",
        node_type="server",
        context=context,
        access_point="tcp://*:5555",
    )
    for sock_addr in [socket_address, socket.getsockopt_string(zmq.LAST_ENDPOINT)]:
        assert sock_addr.startswith("tcp://")
        assert sock_addr.endswith(":5555")
    socket.close()

    socket, socket_address = BaseZMQ.get_socket(
        server_id="test-server",
        socket_id="test-server",
        node_type="server",
        context=context,
        access_point="IPC",
    )

    assert socket_address == socket.getsockopt_string(zmq.LAST_ENDPOINT)
    assert socket_address.startswith("ipc://")
    assert socket_address.endswith(".ipc")
    socket.close()

    socket, socket_address = BaseZMQ.get_socket(
        server_id="test-server",
        socket_id="test-server",
        node_type="server",
        context=context,
        access_point="INPROC",
    )
    assert socket_address == socket.getsockopt_string(zmq.LAST_ENDPOINT)
    assert socket_address.startswith("inproc://")
    assert socket_address.endswith("test-server")
    socket.close()
    context.term()

    # Specify transport as enum and do the same tests
    context = zmq.Context()
    socket, socket_address = BaseZMQ.get_socket(
        server_id="test-server",
        socket_id="test-server",
        node_type="server",
        context=context,
        access_point=ZMQ_TRANSPORTS.INPROC,
    )
    assert socket_address.startswith("inproc://")
    assert socket_address.endswith("test-server")
    socket.close()

    socket, socket_address = BaseZMQ.get_socket(
        server_id="test-server",
        socket_id="test-server",
        node_type="server",
        context=context,
        access_point=ZMQ_TRANSPORTS.IPC,
    )
    assert socket_address.startswith("ipc://")
    assert socket_address.endswith(".ipc")
    socket.close()

    socket, socket_address = BaseZMQ.get_socket(
        server_id="test-server",
        socket_id="test-server",
        node_type="server",
        context=context,
        access_point=ZMQ_TRANSPORTS.TCP,
    )
    assert socket_address.startswith("tcp://")
    # Strip the port number from TCP address and check if it's a valid port integer
    host, port_str = socket_address.rsplit(":", 1)
    assert port_str.isdigit()
    assert 0 < int(port_str) < 65536
    socket.close()
    context.term()

    # check that other transport options raise error
    context = zmq.asyncio.Context()
    with pytest.raises(NotImplementedError):
        BaseZMQ.get_socket(
            server_id="test-server",
            socket_id="test-server",
            node_type="server",
            context=context,
            access_point="PUB",
        )


def test_04_socket_options():
    """check that socket options are as expected"""
    context = zmq.asyncio.Context()

    socket, _ = BaseZMQ.get_socket(
        server_id="test-server",
        socket_id="test-server",
        node_type="server",
        context=context,
        socket_type=zmq.ROUTER,
    )
    assert socket.socket_type == zmq.ROUTER
    assert socket.getsockopt_string(zmq.IDENTITY) == "test-server"
    assert isinstance(socket, zmq.asyncio.Socket)
    socket.close()

    socket, _ = BaseZMQ.get_socket(
        server_id="test-server",
        socket_id="test-server",
        node_type="server",
        context=context,
        socket_type=zmq.DEALER,
    )
    assert socket.socket_type == zmq.DEALER
    assert socket.getsockopt_string(zmq.IDENTITY) == "test-server"
    assert isinstance(socket, zmq.asyncio.Socket)
    socket.close()

    socket, _ = BaseZMQ.get_socket(
        server_id="test-server",
        socket_id="test-server",
        node_type="server",
        context=context,
        socket_type=zmq.PUB,
    )
    assert socket.socket_type == zmq.PUB
    assert socket.getsockopt_string(zmq.IDENTITY) == "test-server"
    assert isinstance(socket, zmq.asyncio.Socket)
    socket.close()

    socket, _ = BaseZMQ.get_socket(
        server_id="test-server",
        socket_id="test-server",
        node_type="server",
        context=context,
        socket_type=zmq.SUB,
    )
    assert socket.socket_type == zmq.SUB
    assert socket.getsockopt_string(zmq.IDENTITY) == "test-server"
    assert isinstance(socket, zmq.asyncio.Socket)
    socket.close()

    socket, _ = BaseZMQ.get_socket(
        server_id="test-server",
        socket_id="test-server",
        node_type="server",
        context=context,
        socket_type=zmq.PAIR,
    )
    assert socket.socket_type == zmq.PAIR
    assert socket.getsockopt_string(zmq.IDENTITY) == "test-server"
    assert isinstance(socket, zmq.asyncio.Socket)
    socket.close()

    socket, _ = BaseZMQ.get_socket(
        server_id="test-server",
        socket_id="test-server",
        node_type="server",
        context=context,
        socket_type=zmq.PUSH,
    )
    assert socket.socket_type == zmq.PUSH
    assert socket.getsockopt_string(zmq.IDENTITY) == "test-server"
    assert isinstance(socket, zmq.asyncio.Socket)
    socket.close()

    socket, _ = BaseZMQ.get_socket(
        server_id="test-server",
        socket_id="test-server",
        node_type="server",
        context=context,
        socket_type=zmq.PULL,
    )
    assert socket.socket_type == zmq.PULL
    assert socket.getsockopt_string(zmq.IDENTITY) == "test-server"
    assert isinstance(socket, zmq.asyncio.Socket)
    socket.close()
    context.term()


"""
TODO:
1. check node_type values
2. check if TCP socket search happens
"""

if __name__ == "__main__":
    pytest.main([__file__, "-v", "-s"])



import asyncio

from copy import deepcopy

import pytest

from hololinked.core.actions import (
    Action,
    BoundAction,
    BoundAsyncAction,
    BoundSyncAction,
)
from hololinked.core.dataklasses import ActionInfoValidator
from hololinked.core.thing import action
from hololinked.schema_validators import JSONSchemaValidator
from hololinked.td.interaction_affordance import ActionAffordance
from hololinked.utils import isclassmethod


try:
    from .things import TestThing
    from .things.test_thing import replace_methods_with_actions
except ImportError:
    from things import TestThing
    from things.test_thing import replace_methods_with_actions


@pytest.fixture(scope="module")
def thing() -> TestThing:
    thing_cls = deepcopy(TestThing)
    _thing = thing_cls(id="test-action")
    replace_methods_with_actions(thing_cls=thing_cls)
    return _thing


def test_01_allowed_actions():
    """Test if methods can be decorated with action"""
    # 1. instance method can be decorated with action
    assert TestThing.action_echo == action()(TestThing.action_echo.obj)  # already predecorated as action
    # 2. classmethod can be decorated with action
    assert Action(TestThing.action_echo_with_classmethod) == action()(TestThing.action_echo_with_classmethod)
    assert isclassmethod(TestThing.action_echo_with_classmethod)
    # 3. async methods can be decorated with action
    assert Action(TestThing.action_echo_async) == action()(TestThing.action_echo_async)
    # 4. async classmethods can be decorated with action
    assert Action(TestThing.action_echo_async_with_classmethod) == action()(
        TestThing.action_echo_async_with_classmethod
    )
    assert isclassmethod(TestThing.action_echo_async_with_classmethod)
    # 5. parameterized function can be decorated with action
    assert Action(TestThing.parameterized_action) == action(safe=True)(TestThing.parameterized_action)
    assert Action(TestThing.parameterized_action_without_call) == action(idempotent=True)(
        TestThing.parameterized_action_without_call
    )
    assert Action(TestThing.parameterized_action_async) == action(synchronous=True)(
        TestThing.parameterized_action_async
    )
    # 6. actions with input and output schema
    assert Action(TestThing.json_schema_validated_action) == action(
        input_schema={
            "val1": "integer",
            "val2": "string",
            "val3": "object",
            "val4": "array",
        },
        output_schema={"val1": "int", "val3": "dict"},
    )(TestThing.json_schema_validated_action)
    assert Action(TestThing.pydantic_validated_action) == action()(TestThing.pydantic_validated_action)


def test_02_bound_method(thing: TestThing):
    """Test if methods decorated with action are correctly bound"""
    # 1. instance method can be decorated with action
    assert isinstance(thing.action_echo, BoundAction)
    assert isinstance(thing.action_echo, BoundSyncAction)
    assert not isinstance(thing.action_echo, BoundAsyncAction)
    assert isinstance(TestThing.action_echo, Action)
    assert not isinstance(TestThing.action_echo, BoundAction)
    # associated attributes of BoundAction
    assert isinstance(thing.action_echo, BoundAction)
    assert thing.action_echo.name == "action_echo"
    assert thing.action_echo.owner_inst == thing
    assert thing.action_echo.owner == TestThing
    assert thing.action_echo.execution_info == TestThing.action_echo.execution_info
    assert str(thing.action_echo) == f"<BoundAction({TestThing.__name__}.{thing.action_echo.name} of {thing.id})>"
    assert thing.action_echo != TestThing.action_echo
    assert thing.action_echo.bound_obj == thing

    # 2. classmethod can be decorated with action
    assert isinstance(thing.action_echo_with_classmethod, BoundAction)
    assert isinstance(thing.action_echo_with_classmethod, BoundSyncAction)
    assert not isinstance(thing.action_echo_with_classmethod, BoundAsyncAction)
    assert isinstance(TestThing.action_echo_with_classmethod, BoundAction)
    assert isinstance(TestThing.action_echo_with_classmethod, BoundSyncAction)
    assert not isinstance(TestThing.action_echo_with_classmethod, Action)
    # associated attributes of BoundAction
    assert isinstance(thing.action_echo_with_classmethod, BoundAction)
    assert thing.action_echo_with_classmethod.name == "action_echo_with_classmethod"
    assert thing.action_echo_with_classmethod.owner_inst == thing
    assert thing.action_echo_with_classmethod.owner == TestThing
    assert thing.action_echo_with_classmethod.execution_info == TestThing.action_echo_with_classmethod.execution_info
    assert (
        str(thing.action_echo_with_classmethod)
        == f"<BoundAction({TestThing.__name__}.{thing.action_echo_with_classmethod.name} of {thing.id})>"
    )
    assert thing.action_echo_with_classmethod == TestThing.action_echo_with_classmethod
    assert thing.action_echo_with_classmethod.bound_obj == TestThing

    # 3. async methods can be decorated with action
    assert isinstance(thing.action_echo_async, BoundAction)
    assert not isinstance(thing.action_echo_async, BoundSyncAction)
    assert isinstance(thing.action_echo_async, BoundAsyncAction)
    assert isinstance(TestThing.action_echo_async, Action)
    assert not isinstance(TestThing.action_echo_async, BoundAction)
    # associated attributes of BoundAction
    assert isinstance(thing.action_echo_async, BoundAction)
    assert thing.action_echo_async.name == "action_echo_async"
    assert thing.action_echo_async.owner_inst == thing
    assert thing.action_echo_async.owner == TestThing
    assert thing.action_echo_async.execution_info == TestThing.action_echo_async.execution_info
    assert (
        str(thing.action_echo_async)
        == f"<BoundAction({TestThing.__name__}.{thing.action_echo_async.name} of {thing.id})>"
    )
    assert thing.action_echo_async != TestThing.action_echo_async
    assert thing.action_echo_async.bound_obj == thing

    # 4. async classmethods can be decorated with action
    assert isinstance(thing.action_echo_async_with_classmethod, BoundAction)
    assert not isinstance(thing.action_echo_async_with_classmethod, BoundSyncAction)
    assert isinstance(thing.action_echo_async_with_classmethod, BoundAsyncAction)
    assert isinstance(TestThing.action_echo_async_with_classmethod, BoundAction)
    assert isinstance(TestThing.action_echo_async_with_classmethod, BoundAsyncAction)
    assert not isinstance(TestThing.action_echo_async_with_classmethod, Action)
    # associated attributes of BoundAction
    assert isinstance(thing.action_echo_async_with_classmethod, BoundAction)
    assert thing.action_echo_async_with_classmethod.name == "action_echo_async_with_classmethod"
    assert thing.action_echo_async_with_classmethod.owner_inst == thing
    assert thing.action_echo_async_with_classmethod.owner == TestThing
    assert (
        thing.action_echo_async_with_classmethod.execution_info
        == TestThing.action_echo_async_with_classmethod.execution_info
    )
    assert (
        str(thing.action_echo_async_with_classmethod)
        == f"<BoundAction({TestThing.__name__}.{thing.action_echo_async_with_classmethod.name} of {thing.id})>"
    )
    assert thing.action_echo_async_with_classmethod == TestThing.action_echo_async_with_classmethod
    assert thing.action_echo_async_with_classmethod.bound_obj == TestThing

    # 5. parameterized function can be decorated with action
    assert isinstance(thing.parameterized_action, BoundAction)
    assert isinstance(thing.parameterized_action, BoundSyncAction)
    assert not isinstance(thing.parameterized_action, BoundAsyncAction)
    assert isinstance(TestThing.parameterized_action, Action)
    assert not isinstance(TestThing.parameterized_action, BoundAction)
    # associated attributes of BoundAction
    assert isinstance(thing.parameterized_action, BoundAction)
    assert thing.parameterized_action.name == "parameterized_action"
    assert thing.parameterized_action.owner_inst == thing
    assert thing.parameterized_action.owner == TestThing
    assert thing.parameterized_action.execution_info == TestThing.parameterized_action.execution_info
    assert (
        str(thing.parameterized_action)
        == f"<BoundAction({TestThing.__name__}.{thing.parameterized_action.name} of {thing.id})>"
    )
    assert thing.parameterized_action != TestThing.parameterized_action
    assert thing.parameterized_action.bound_obj == thing

    # 6. parameterized function can be decorated with action
    assert isinstance(thing.parameterized_action_without_call, BoundAction)
    assert isinstance(thing.parameterized_action_without_call, BoundSyncAction)
    assert not isinstance(thing.parameterized_action_without_call, BoundAsyncAction)
    assert isinstance(TestThing.parameterized_action_without_call, Action)
    assert not isinstance(TestThing.parameterized_action_without_call, BoundAction)
    # associated attributes of BoundAction
    assert isinstance(thing.parameterized_action_without_call, BoundAction)
    assert thing.parameterized_action_without_call.name == "parameterized_action_without_call"
    assert thing.parameterized_action_without_call.owner_inst == thing
    assert thing.parameterized_action_without_call.owner == TestThing
    assert (
        thing.parameterized_action_without_call.execution_info
        == TestThing.parameterized_action_without_call.execution_info
    )
    assert (
        str(thing.parameterized_action_without_call)
        == f"<BoundAction({TestThing.__name__}.{thing.parameterized_action_without_call.name} of {thing.id})>"
    )
    assert thing.parameterized_action_without_call != TestThing.parameterized_action_without_call
    assert thing.parameterized_action_without_call.bound_obj == thing

    # 7. parameterized function can be decorated with action
    assert isinstance(thing.parameterized_action_async, BoundAction)
    assert not isinstance(thing.parameterized_action_async, BoundSyncAction)
    assert isinstance(thing.parameterized_action_async, BoundAsyncAction)
    assert isinstance(TestThing.parameterized_action_async, Action)
    assert not isinstance(TestThing.parameterized_action_async, BoundAction)
    # associated attributes of BoundAction
    assert isinstance(thing.parameterized_action_async, BoundAction)
    assert thing.parameterized_action_async.name == "parameterized_action_async"
    assert thing.parameterized_action_async.owner_inst == thing
    assert thing.parameterized_action_async.owner == TestThing
    assert thing.parameterized_action_async.execution_info == TestThing.parameterized_action_async.execution_info
    assert (
        str(thing.parameterized_action_async)
        == f"<BoundAction({TestThing.__name__}.{thing.parameterized_action_async.name} of {thing.id})>"
    )
    assert thing.parameterized_action_async != TestThing.parameterized_action_async
    assert thing.parameterized_action_async.bound_obj == thing

    # 8. actions with input and output schema
    assert isinstance(thing.json_schema_validated_action, BoundAction)
    assert isinstance(thing.json_schema_validated_action, BoundSyncAction)
    assert not isinstance(thing.json_schema_validated_action, BoundAsyncAction)
    assert isinstance(TestThing.json_schema_validated_action, Action)
    assert not isinstance(TestThing.json_schema_validated_action, BoundAction)
    # associated attributes of BoundAction
    assert isinstance(thing.json_schema_validated_action, BoundAction)
    assert thing.json_schema_validated_action.name == "json_schema_validated_action"
    assert thing.json_schema_validated_action.owner_inst == thing
    assert thing.json_schema_validated_action.owner == TestThing
    assert thing.json_schema_validated_action.execution_info == TestThing.json_schema_validated_action.execution_info
    assert (
        str(thing.json_schema_validated_action)
        == f"<BoundAction({TestThing.__name__}.{thing.json_schema_validated_action.name} of {thing.id})>"
    )
    assert thing.json_schema_validated_action != TestThing.json_schema_validated_action
    assert thing.json_schema_validated_action.bound_obj == thing


def test_03_remote_info():
    """Test if the validator is working correctly, on which the logic of the action is based"""
    remote_info = TestThing.action_echo.execution_info
    assert isinstance(remote_info, ActionInfoValidator)
    assert remote_info.isaction
    assert not remote_info.isproperty
    assert not remote_info.isparameterized
    assert not remote_info.iscoroutine
    assert not remote_info.safe
    assert not remote_info.idempotent
    assert remote_info.synchronous

    remote_info = TestThing.action_echo_async.execution_info
    assert isinstance(remote_info, ActionInfoValidator)
    assert remote_info.isaction
    assert remote_info.iscoroutine
    assert not remote_info.isproperty
    assert not remote_info.isparameterized
    assert not remote_info.safe
    assert not remote_info.idempotent
    assert remote_info.synchronous

    remote_info = TestThing.action_echo_with_classmethod.execution_info
    assert isinstance(remote_info, ActionInfoValidator)
    assert remote_info.isaction
    assert not remote_info.iscoroutine
    assert not remote_info.isproperty
    assert not remote_info.isparameterized
    assert not remote_info.safe
    assert not remote_info.idempotent
    assert remote_info.synchronous

    remote_info = TestThing.parameterized_action.execution_info
    assert isinstance(remote_info, ActionInfoValidator)
    assert remote_info.isaction
    assert not remote_info.iscoroutine
    assert not remote_info.isproperty
    assert remote_info.isparameterized
    assert remote_info.safe
    assert not remote_info.idempotent
    assert remote_info.synchronous

    remote_info = TestThing.parameterized_action_without_call.execution_info
    assert isinstance(remote_info, ActionInfoValidator)
    assert remote_info.isaction
    assert not remote_info.iscoroutine
    assert not remote_info.isproperty
    assert remote_info.isparameterized
    assert not remote_info.safe
    assert remote_info.idempotent
    assert remote_info.synchronous

    remote_info = TestThing.parameterized_action_async.execution_info
    assert isinstance(remote_info, ActionInfoValidator)
    assert remote_info.isaction
    assert remote_info.iscoroutine
    assert not remote_info.isproperty
    assert remote_info.isparameterized
    assert not remote_info.safe
    assert not remote_info.idempotent
    assert remote_info.synchronous

    remote_info = TestThing.json_schema_validated_action.execution_info
    assert isinstance(remote_info, ActionInfoValidator)
    assert remote_info.isaction
    assert not remote_info.iscoroutine
    assert not remote_info.isproperty
    assert not remote_info.isparameterized
    assert not remote_info.safe
    assert not remote_info.idempotent
    assert remote_info.synchronous
    assert isinstance(remote_info.schema_validator, JSONSchemaValidator)


def test_04_api_and_invalid_actions():
    """Test if action prevents invalid objects from being named as actions and raises neat errors"""
    # done allow action decorator to be terminated without '()' on a method
    with pytest.raises(TypeError) as ex:
        action(TestThing.incorrectly_decorated_method)
    assert str(ex.value).startswith(
        "input schema should be a JSON or pydantic BaseModel, not a function/method, did you decorate your action wrongly?"
    )

    # dunder methods cannot be decorated with action
    with pytest.raises(ValueError) as ex:
        action()(TestThing.__internal__)
    assert str(ex.value).startswith("dunder objects cannot become remote")

    # only functions and methods can be decorated with action
    for obj in [
        TestThing,
        str,
        1,
        1.0,
        "Str",
        True,
        None,
        object(),
        type,
        property,
    ]:
        with pytest.raises(TypeError) as ex2:
            action()(obj)
        assert str(ex2.value).startswith("target for action or is not a function/method.")

    with pytest.raises(ValueError) as ex:
        action(safe=True, some_kw=1)
    assert str(ex.value).startswith("Only 'safe', 'idempotent', 'synchronous' are allowed")


def test_05_thing_cls_actions(thing: TestThing):
    """Test class and instance level action access"""
    # class level
    for name, act in TestThing.actions.descriptors.items():
        assert isinstance(act, Action)
    for name in replace_methods_with_actions._exposed_actions:
        assert name in TestThing.actions
    # instance level
    for name, act in thing.actions.values.items():
        assert isinstance(act, BoundAction)
    for name in replace_methods_with_actions._exposed_actions:
        assert name in thing.actions
    # cannot call an instance bound action at class level
    with pytest.raises(NotImplementedError):
        TestThing.action_echo(thing, 1)
    # but can call instance bound action with instance
    assert thing.action_echo(1) == 1
    # can also call classmethods as usual
    assert TestThing.action_echo_with_classmethod(2) == 2
    assert thing.action_echo_with_classmethod(3) == 3
    # async methods behave similarly
    assert asyncio.run(thing.action_echo_async(4)) == 4
    assert asyncio.run(TestThing.action_echo_async_with_classmethod(5)) == 5
    with pytest.raises(NotImplementedError):
        asyncio.run(TestThing.action_echo(7))
    # parameterized actions behave similarly
    assert thing.parameterized_action(1, "hello1", 1.1) == ("test-action", 1, "hello1", 1.1)
    assert asyncio.run(thing.parameterized_action_async(2, "hello2", "foo2")) == ("test-action", 2, "hello2", "foo2")
    with pytest.raises(NotImplementedError):
        TestThing.parameterized_action(3, "hello3", 5)
    with pytest.raises(NotImplementedError):
        asyncio.run(TestThing.parameterized_action_async(4, "hello4", 5))


def test_06_action_affordance(thing: TestThing):
    """Test if action affordance is correctly created"""
    assert isinstance(thing.action_echo, BoundAction)
    affordance = thing.action_echo.to_affordance()
    assert isinstance(affordance, ActionAffordance)
    assert affordance.idempotent is None
    assert affordance.synchronous is True
    assert affordance.safe is None
    assert affordance.input is None
    assert affordance.output is None
    assert affordance.description is None

    assert isinstance(thing.action_echo_with_classmethod, BoundAction)
    affordance = thing.action_echo_with_classmethod.to_affordance()
    assert isinstance(affordance, ActionAffordance)
    assert affordance.idempotent is None
    assert affordance.synchronous is True
    assert affordance.safe is None
    assert affordance.input is None
    assert affordance.output is None
    assert affordance.description is None

    assert isinstance(thing.action_echo_async, BoundAction)
    affordance = thing.action_echo_async.to_affordance()
    assert isinstance(affordance, ActionAffordance)
    assert affordance.idempotent is None
    assert affordance.synchronous is True
    assert affordance.safe is None
    assert affordance.input is None
    assert affordance.output is None
    assert affordance.description is None

    assert isinstance(thing.action_echo_async_with_classmethod, BoundAction)
    affordance = thing.action_echo_async_with_classmethod.to_affordance()
    assert isinstance(affordance, ActionAffordance)
    assert affordance.idempotent is None
    assert affordance.synchronous is True
    assert affordance.safe is None
    assert affordance.input is None
    assert affordance.output is None
    assert affordance.description is None

    assert isinstance(thing.parameterized_action, BoundAction)
    affordance = thing.parameterized_action.to_affordance()
    assert isinstance(affordance, ActionAffordance)
    assert affordance.idempotent is None
    assert affordance.synchronous is True
    assert affordance.safe is True
    assert affordance.input is None
    assert affordance.output is None
    assert affordance.description is None

    assert isinstance(thing.parameterized_action_without_call, BoundAction)
    affordance = thing.parameterized_action_without_call.to_affordance()
    assert isinstance(affordance, ActionAffordance)
    assert affordance.idempotent is True
    assert affordance.synchronous is True
    assert affordance.safe is None
    assert affordance.input is None
    assert affordance.output is None
    assert affordance.description is None

    assert isinstance(thing.parameterized_action_async, BoundAction)
    affordance = thing.parameterized_action_async.to_affordance()
    assert isinstance(affordance, ActionAffordance)
    assert affordance.idempotent is None
    assert affordance.synchronous is True
    assert affordance.safe is None
    assert affordance.input is None
    assert affordance.output is None
    assert affordance.description is None

    assert isinstance(thing.json_schema_validated_action, BoundAction)
    affordance = thing.json_schema_validated_action.to_affordance()
    assert isinstance(affordance, ActionAffordance)
    assert affordance.idempotent is None
    assert affordance.synchronous is True
    assert affordance.safe is None
    assert isinstance(affordance.input, dict)
    assert isinstance(affordance.output, dict)
    assert affordance.description is None


if __name__ == "__main__":
    pytest.main([__file__, "-v", "-s"])



import base64
import os
import threading

from datetime import datetime
from sqlite3 import DatabaseError
from typing import Any

from pymongo import MongoClient
from pymongo import errors as mongo_errors
from sqlalchemy import create_engine, select
from sqlalchemy import inspect as inspect_database
from sqlalchemy.orm import sessionmaker

from ..config import global_config
from ..core.property import Property
from ..param import Parameterized
from ..serializers.serializers import PythonBuiltinJSONSerializer as JSONSerializer
from ..serializers.serializers import Serializers
from .config import MongoDBConfig, SQLDBConfig, SQLiteConfig
from .models import SerializedProperty, ThingInformation, ThingTableBase
from .utils import get_sanitized_filename_from_thing_instance


class BaseDB:
    """
    Database base class irrespective of sync or async implementation.
    Implements configuration file reader.
    """

    def __init__(self, instance: Parameterized, config_file: str | None = None) -> None:
        self.thing_instance = instance
        self.conf = BaseDB.load_conf(
            config_file,
            default_file_path=os.path.join(
                global_config.TEMP_DIR_DB,
                get_sanitized_filename_from_thing_instance(instance),
            ),
        )
        self.URL = self.conf.URL
        self._batch_call_context = {}

    @classmethod
    def load_conf(
        cls,
        config_file: str,
        default_file_path: str = "",
    ) -> SQLDBConfig | SQLiteConfig | MongoDBConfig:
        """
        load configuration file using JSON serializer
        """
        if not config_file:
            return SQLiteConfig(file=default_file_path)
        elif config_file.endswith(".json"):
            file = open(config_file, "r")
            conf = JSONSerializer.load(file)
            if conf.get("provider", None) in ["postgresql", "mysql"]:
                return SQLDBConfig.model_validate(conf, strict=True, from_attributes=True)
            elif conf.get("provider", None) == "sqlite":
                return SQLiteConfig.model_validate(conf, strict=True, from_attributes=True)
            elif conf.get("provider", None) == "mongo":
                return MongoDBConfig.model_validate(conf, strict=True, from_attributes=True)
            raise NotImplementedError("only postgresql, mysql, sqlite and mongo are supported")
        raise ValueError("config files of extension {} expected, given file name {}".format(["json"], config_file))

    @property
    def in_batch_call_context(self):
        return threading.get_ident() in self._batch_call_context


"""
class BaseAsyncDB(BaseDB):
    Base class for an async database engine, creates sqlalchemy engine & session.

    This class is not fully implemented yet. 
    
    Set `async_db_engine` boolean flag to True in `Thing` class to use this engine.
    Database operations are then scheduled in the event loop instead of blocking the current thread.
    Scheduling happens after properties are set/written.

    Parameters
    ----------
    database: str
        The database to open in the database server specified in config_file (see below)
    serializer: BaseSerializer
        The serializer to use for serializing and deserializing data (for example
        property serializing before writing to database). Will be the same as zmq_serializer supplied to `Thing`.
    config_file: str
        absolute path to database server configuration file
    
    def __init__(
        self,
        instance: Parameterized,
        serializer: BaseSerializer | None = None,
        config_file: str | None = None,
    ) -> None:
        super().__init__(instance=instance, serializer=serializer, config_file=config_file)
        self.engine = asyncio_ext.create_async_engine(self.URL)
        self.async_session = sessionmaker(self.engine, expire_on_commit=True, class_=asyncio_ext.AsyncSession)
        if self.conf.provider == "sqlite":
            ThingTableBase.metadata.create_all(self.engine)
"""


class BaseSyncDB(BaseDB):
    """
    Base class for a synchronous (blocking) database engine, implements sqlalchemy engine & session creation.
    Default DB engine for `Thing` & called immediately after properties are set/written.

    Parameters
    ----------
    instance: Parameterized
        The `Thing` instance to which this database engine belongs
    config_file: str
        path to database server configuration file
    """

    def __init__(self, instance: Parameterized, config_file: str | None = None) -> None:
        super().__init__(instance=instance, config_file=config_file)
        self.engine = create_engine(self.URL)
        self.sync_session = sessionmaker(self.engine, expire_on_commit=True)
        if self.conf.provider == "sqlite":
            ThingTableBase.metadata.create_all(self.engine)


class ThingDB(BaseSyncDB):
    """
    Synchronous database engine composed within `Thing`.
    Carries out database operations like storing object information, properties etc.
    """

    def fetch_own_info(self):  # -> ThingInformation:
        """
        fetch `Thing` instance's own information (some useful metadata which helps the `Thing` run).

        Returns
        -------
        `ThingInformation`
        """
        if not inspect_database(self.engine).has_table("things"):
            return
        with self.sync_session() as session:
            stmt = select(ThingInformation).filter_by(
                thing_id=self.thing_instance.id,
                thing_class=self.thing_instance.__class__.__name__,
            )
            data = session.execute(stmt)
            data = data.scalars().all()
            if len(data) == 0:
                return None
            elif len(data) == 1:
                return data[0]
            else:
                raise DatabaseError(
                    "Multiple things with same instance name found, either cleanup database/detach/make new"
                )

    def get_property(self, property: str | Property, deserialized: bool = True) -> Any:
        """
        Fetch a single property.

        Parameters
        ----------
        property: str | Property
            string name or descriptor object
        deserialized: bool, default True
            deserialize the property if True

        Returns
        -------
        Any
            property value
        """
        with self.sync_session() as session:
            name = property if isinstance(property, str) else property.name
            stmt = select(SerializedProperty).filter_by(
                thing_id=self.thing_instance.id,
                thing_class=self.thing_instance.__class__.__name__,
                name=name,
            )
            data = session.execute(stmt)
            prop = data.scalars().all()  # type: list[SerializedProperty]
            if len(prop) == 0:
                raise DatabaseError(f"property {name} not found in database")
            elif len(prop) > 1:
                raise DatabaseError("multiple properties with same name found")  # Impossible actually
            if not deserialized:
                return prop[0]
            serializer = Serializers.content_types.get(prop[0].content_type, None) or Serializers.for_object(
                self.thing_instance.id, self.thing_instance.__class__.__name__, name
            )
            return serializer.loads(prop[0].serialized_value)

    def set_property(self, property: str | Property, value: Any) -> None:
        """
        Change the value of an already existing property.

        Parameters
        ----------
        property: str | Property
            string name or descriptor object
        value: Any
            value of the property
        """
        if self.in_batch_call_context:
            self._batch_call_context[threading.get_ident()][property.name] = value
            return
        with self.sync_session() as session:
            name = property if isinstance(property, str) else property.name
            stmt = select(SerializedProperty).filter_by(
                thing_id=self.thing_instance.id,
                thing_class=self.thing_instance.__class__.__name__,
                name=name,
            )
            data = session.execute(stmt)
            prop = data.scalars().all()
            if len(prop) > 1:
                raise DatabaseError("multiple properties with same name found")  # Impossible actually
            if len(prop) == 1:
                prop = prop[0]
                serializer = Serializers.content_types.get(prop.content_type, None) or Serializers.for_object(
                    self.thing_instance.id, self.thing_instance.__class__.__name__, name
                )
                prop.serialized_value = serializer.dumps(value)
                prop.updated_at = datetime.now().isoformat()
                prop.content_type = serializer.content_type
            else:
                serializer = Serializers.for_object(
                    self.thing_instance.id, self.thing_instance.__class__.__name__, name
                )
                now = datetime.now().isoformat()
                prop = SerializedProperty(
                    id=None,
                    name=name,
                    serialized_value=serializer.dumps(value),
                    thing_id=self.thing_instance.id,
                    thing_class=self.thing_instance.__class__.__name__,
                    created_at=now,
                    updated_at=now,
                    content_type=serializer.content_type,
                )
                session.add(prop)
            session.commit()

    def get_properties(self, properties: dict[str | Property, Any], deserialized: bool = True) -> dict[str, Any]:
        """
        Get multiple properties at once.

        Parameters
        ----------
        properties: List[str | Property]
            string names or the descriptor of the properties as a list
        deserialized: bool, default True
            deserilize the properties if True

        Returns
        -------
        dict[str, Any]
            property names and values as items
        """
        with self.sync_session() as session:
            names = []
            for obj in properties.keys():
                names.append(obj if isinstance(obj, str) else obj.name)
            stmt = (
                select(SerializedProperty)
                .filter_by(thing_id=self.thing_instance.id, thing_class=self.thing_instance.__class__.__name__)
                .filter(SerializedProperty.name.in_(names))
            )
            data = session.execute(stmt)
            unserialized_props = data.scalars().all()
            props = dict()
            for prop in unserialized_props:
                serializer = Serializers.content_types.get(prop.content_type, None) or Serializers.for_object(
                    self.thing_instance.id,
                    self.thing_instance.__class__.__name__,
                    prop.name,
                )
                props[prop.name] = (
                    prop.serialized_value if not deserialized else serializer.loads(prop.serialized_value)
                )
            return props

    def set_properties(self, properties: dict[str | Property, Any]) -> None:
        """
        Change the values of already existing properties at once

        Parameters
        ----------
        properties: Dict[str | Property, Any]
            string names or the descriptor of the property and any value as dictionary pairs
        """
        if self.in_batch_call_context:
            for obj, value in properties.items():
                name = obj if isinstance(obj, str) else obj.name
                self._batch_call_context[threading.get_ident()][name] = value
            return
        with self.sync_session() as session:
            names = []
            for obj in properties.keys():
                names.append(obj if isinstance(obj, str) else obj.name)
            stmt = (
                select(SerializedProperty)
                .filter_by(thing_id=self.thing_instance.id, thing_class=self.thing_instance.__class__.__name__)
                .filter(SerializedProperty.name.in_(names))
            )
            data = session.execute(stmt)
            db_props = data.scalars().all()
            for obj, value in properties.items():
                name = obj if isinstance(obj, str) else obj.name
                db_prop = list(filter(lambda db_prop: db_prop.name == name, db_props))  # type: list[SerializedProperty]
                if len(db_prop) > 1:
                    raise DatabaseError("multiple properties with same name found")  # Impossible actually
                if len(db_prop) == 1:
                    db_prop = db_prop[0]  # type: SerializedProperty
                    serializer = Serializers.content_types.get(db_prop.content_type, None) or Serializers.for_object(
                        self.thing_instance.id, self.thing_instance.__class__.__name__, name
                    )
                    db_prop.serialized_value = serializer.dumps(value)
                    db_prop.updated_at = datetime.now().isoformat()
                    db_prop.content_type = serializer.content_type
                else:
                    serializer = Serializers.for_object(
                        self.thing_instance.id, self.thing_instance.__class__.__name__, name
                    )
                    now = datetime.now().isoformat()
                    prop = SerializedProperty(
                        id=None,
                        name=name,
                        serialized_value=serializer.dumps(value),
                        thing_id=self.thing_instance.id,
                        thing_class=self.thing_instance.__class__.__name__,
                        created_at=now,
                        updated_at=now,
                        content_type=serializer.content_type,
                    )
                    session.add(prop)
            session.commit()

    def get_all_properties(self, deserialized: bool = True) -> dict[str, Any]:
        """
        Get all properties of the `Thing` instance.

        Parameters
        ----------
        deserialized: bool, default True
            deserilize the properties if True

        Returns
        -------
        dict[str, Any]
            property names and values as items
        """
        with self.sync_session() as session:
            stmt = select(SerializedProperty).filter_by(
                thing_id=self.thing_instance.id, thing_class=self.thing_instance.__class__.__name__
            )
            data = session.execute(stmt)
            existing_props = data.scalars().all()  # type: list[SerializedProperty]
            if not deserialized:
                return existing_props
            props = dict()
            for prop in existing_props:
                serializer = Serializers.content_types.get(prop.content_type, None) or Serializers.for_object(
                    self.thing_instance.id, self.thing_instance.__class__.__name__, prop.name
                )
                props[prop.name] = serializer.loads(prop.serialized_value)
            return props

    def create_missing_properties(
        self,
        properties: dict[str, Property],
        get_missing_property_names: bool = False,
    ) -> None | list[str]:
        """
        Create any and all missing properties of `Thing` instance in database.

        Parameters
        ----------
        properties: Dict[str, Property]
            descriptors of the properties
        get_missing_property_names: bool, default False
            whether to return the list of missing property names

        Returns
        -------
        List[str]
            list of missing properties if get_missing_property_names is True
        """
        missing_props = []
        with self.sync_session() as session:
            existing_props = self.get_all_properties()
            for prop in properties.values():
                if prop.name not in existing_props:
                    serializer = Serializers.for_object(
                        self.thing_instance.id, self.thing_instance.__class__.__name__, prop.name
                    )
                    now = datetime.now().isoformat()
                    prop = SerializedProperty(
                        id=None,
                        name=prop.name,
                        serialized_value=serializer.dumps(getattr(self.thing_instance, prop.name)),
                        thing_id=self.thing_instance.id,
                        thing_class=self.thing_instance.__class__.__name__,
                        created_at=now,
                        updated_at=now,
                        content_type=serializer.content_type,
                    )
                    session.add(prop)
                    missing_props.append(prop.name)
            session.commit()
        if get_missing_property_names:
            return missing_props

    def create_db_init_properties(self, thing_id: str = None, thing_class: str = None, **properties: Any) -> None:
        """
        Create properties that are supposed to be initialized from database for a thing instance.
        Invoke this method once before running the thing instance to store its initial value in database.

        Parameters
        ----------
        thing_id: str
            ID of the thing instance to which these properties belong
        thing_class: str
            Class name of the thing instance to which these properties belong
        properties: dict[str, Any]
            property names and their initial values as dictionary pairs
        """
        with self.sync_session() as session:
            for name, value in properties.items():
                serializer = Serializers.for_object(thing_id, thing_class, name)
                now = datetime.now().isoformat()
                prop = SerializedProperty(
                    id=None,
                    name=name,
                    serialized_value=serializer.dumps(value),
                    thing_id=thing_id,
                    thing_class=thing_class,
                    created_at=now,
                    updated_at=now,
                    content_type=serializer.content_type,
                )
                session.add(prop)
            session.commit()


class batch_db_commit:
    """
    Context manager to write multiple properties to database at once. Useful for sequential sets/writes of multiple properties
    which has db_commit or db_persist set to True, but only write their values to database at once.
    """

    def __init__(self, db_engine: ThingDB) -> None:
        self.db_engine = db_engine

    def __enter__(self) -> None:
        self.db_engine._context[threading.get_ident()] = dict()

    def __exit__(self, exc_type, exc_value, exc_tb) -> None:
        data = self.db_engine._context.pop(threading.get_ident(), dict())  # dict[str, Any]
        if exc_type is None:
            self.db_engine.set_properties(data)
            return
        for name, value in data.items():
            try:
                self.db_engine.set_property(name, value)
            except Exception as ex:
                self.db_engine.thing_instance.logger.error(
                    f"failed to set property {name} to value {value} during batch commit due to exception {ex}"
                )


class MongoThingDB:
    """
    MongoDB-backed database engine for Thing properties and info.

    This class provides persistence for Thing properties using MongoDB.
    Properties are stored in the 'properties' collection, with fields:
    - id: Thing instance identifier
    - name: property name
    - serialized_value: serialized property value

    Methods mirror the interface of ThingDB for compatibility.
    """

    def __init__(self, instance: Parameterized, config_file: str | None = None) -> None:
        """
        Initialize MongoThingDB for a Thing instance.
        Connects to MongoDB and sets up collections.
        """
        self.thing_instance = instance
        self.id = instance.id
        self.config = self.load_conf(config_file)
        self.client = MongoClient(self.config.get("mongo_uri", "mongodb://localhost:27017"))
        self.db = self.client[self.config.get("database", "hololinked")]
        self.properties = self.db["properties"]
        self.things = self.db["things"]

    @classmethod
    def load_conf(cls, config_file: str | None) -> dict[str, Any]:
        """
        Load configuration from JSON file if provided.
        """
        if not config_file:
            return {}
        elif config_file.endswith(".json"):
            with open(config_file, "r") as file:
                return JSONSerializer.load(file)
        else:
            raise ValueError(f"config files of extension - ['json'] expected, given file name {config_file}")

    def fetch_own_info(self):
        """
        Fetch Thing instance metadata from the 'things' collection.
        """
        doc = self.things.find_one({"id": self.id})
        return doc

    def get_property(self, property: str | Property, deserialized: bool = True) -> Any:
        """
        Get a property value from MongoDB for this Thing.
        If deserialized=True, returns the Python value.
        """
        name = property if isinstance(property, str) else property.name
        doc = self.properties.find_one({"id": self.id, "name": name})
        if not doc:
            raise mongo_errors.PyMongoError(f"property {name} not found in database")
        if not deserialized:
            return doc
        serializer = Serializers.for_object(self.id, self.thing_instance.__class__.__name__, name)
        return serializer.loads(base64.b64decode(doc["serialized_value"]))

    def set_property(self, property: str | Property, value: Any) -> None:
        """
        Set a property value in MongoDB for this Thing.
        Value is serialized before storage.
        """
        name = property if isinstance(property, str) else property.name
        serializer = Serializers.for_object(self.id, self.thing_instance.__class__.__name__, name)
        serialized_value = base64.b64encode(serializer.dumps(value)).decode("utf-8")
        self.properties.update_one(
            {"id": self.id, "name": name}, {"$set": {"serialized_value": serialized_value}}, upsert=True
        )

    def get_properties(self, properties: dict[str | Property, Any], deserialized: bool = True) -> dict[str, Any]:
        """
        Get multiple property values from MongoDB for this Thing.
        Returns a dict of property names to values.
        """
        names = [obj if isinstance(obj, str) else obj.name for obj in properties.keys()]
        cursor = self.properties.find({"id": self.id, "name": {"$in": names}})
        result = {}
        for doc in cursor:
            serializer = Serializers.for_object(self.id, self.thing_instance.__class__.__name__, doc["name"])
            result[doc["name"]] = (
                doc["serialized_value"]
                if not deserialized
                else serializer.loads(base64.b64decode(doc["serialized_value"]))
            )
        return result

    def set_properties(self, properties: dict[str | Property, Any]) -> None:
        """
        Set multiple property values in MongoDB for this Thing.
        """
        for obj, value in properties.items():
            name = obj if isinstance(obj, str) else obj.name
            serializer = Serializers.for_object(self.id, self.thing_instance.__class__.__name__, name)
            serialized_value = base64.b64encode(serializer.dumps(value)).decode("utf-8")
            self.properties.update_one(
                {"id": self.id, "name": name}, {"$set": {"serialized_value": serialized_value}}, upsert=True
            )

    def get_all_properties(self, deserialized: bool = True) -> dict[str, Any]:
        cursor = self.properties.find({"id": self.id})
        result = {}
        for doc in cursor:
            serializer = Serializers.for_object(self.id, self.thing_instance.__class__.__name__, doc["name"])
            result[doc["name"]] = (
                doc["serialized_value"]
                if not deserialized
                else serializer.loads(base64.b64decode(doc["serialized_value"]))
            )
        return result

    def create_missing_properties(
        self,
        properties: dict[str, Property],
        get_missing_property_names: bool = False,
    ) -> Any:
        missing_props = []
        existing_props = self.get_all_properties()
        for name, new_prop in properties.items():
            if name not in existing_props:
                serializer = Serializers.for_object(self.id, self.thing_instance.__class__.__name__, new_prop.name)
                serialized_value = base64.b64encode(
                    serializer.dumps(getattr(self.thing_instance, new_prop.name))
                ).decode("utf-8")
                self.properties.insert_one({"id": self.id, "name": new_prop.name, "serialized_value": serialized_value})
                missing_props.append(name)
        if get_missing_property_names:
            return missing_props


__all__ = [
    # BaseAsyncDB.__name__,
    BaseSyncDB.__name__,
    ThingDB.__name__,
    batch_db_commit.__name__,
]



from typing import Any, Callable

import structlog

from .abstractions import ConsumedThingAction, ConsumedThingEvent, ConsumedThingProperty
from .security import APIKeySecurity, BasicSecurity  # noqa: F401


class ObjectProxy:
    """
    Procedural/scripting client for `Thing`. Once connected to a server, properties, methods and events are loaded and
    dynamically populated. Can be used with any supported protocol binding.

    Use `ClientFactory` to create an instance of this class instead of directly creating it.
    """

    _own_attrs = frozenset(
        [
            "__annotations__",
            "_allow_foreign_attributes",
            "id",
            "logger",
            "td",
            "execution_timeout",
            "invokation_timeout",
            "_execution_timeout",
            "_invokation_timeout",
            "_events",
            "_noblock_messages",
            "_schema_validator",
            "_security",
        ]
    )

    __allowed_attribute_types__ = (
        ConsumedThingProperty,
        ConsumedThingAction,
        ConsumedThingEvent,
    )

    def __init__(self, id: str, **kwargs) -> None:
        """
        Parameters
        ----------
        id: str
            unique id for the client
        **kwargs:
            additional keyword arguments:

            - `allow_foreign_attributes`: `bool`, default `False`.
                allows local attributes apart from resources fetched from the server.
            - `logger`: `structlog.stdlib.BoundLogger`, default `None`.
                logger instance
            - `td`: `dict[str, Any]`, default `dict()`.
                Thing Description of the consumed Thing
            - `security`: `BasicSecurity` | `APIKeySecurity`, optional.
                security scheme to be used for authentication
        """
        self.id = id
        self._allow_foreign_attributes = kwargs.get("allow_foreign_attributes", False)
        self._noblock_messages = dict()  # type: dict[str, ConsumedThingAction | ConsumedThingProperty]
        self._schema_validator = kwargs.get("schema_validator", None)
        self._security = kwargs.get("security", None)  # type: BasicSecurity | APIKeySecurity | None
        self.logger = kwargs.pop("logger", structlog.get_logger())
        self.td = kwargs.get("td", dict())  # type: dict[str, Any]

    def __getattribute__(self, __name: str) -> Any:
        obj = super().__getattribute__(__name)
        if isinstance(obj, ConsumedThingProperty):
            return obj.get()
        return obj

    def __setattr__(self, __name: str, __value: Any) -> None:
        if (
            __name in ObjectProxy._own_attrs
            or (__name not in self.__dict__ and isinstance(__value, ObjectProxy.__allowed_attribute_types__))
            or self._allow_foreign_attributes
        ):
            # allowed attribute types are ConsumedThingProperty and ConsumedThingAction defined after this class
            return super(ObjectProxy, self).__setattr__(__name, __value)
        elif __name in self.__dict__:
            obj = self.__dict__[__name]
            if isinstance(obj, ConsumedThingProperty):
                obj.set(value=__value)
                return
            raise AttributeError(f"Cannot set attribute {__name} again to ObjectProxy for {self.id}.")
        raise AttributeError(
            f"Cannot set foreign attribute {__name} to ObjectProxy for {self.id}. Given attribute not found in server object."
        )

    def __repr__(self) -> str:
        return f"ObjectProxy {self.id}"

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        pass

    def __eq__(self, other) -> bool:
        if other is self:
            return True
        return isinstance(other, ObjectProxy) and other.id == self.id and other.TD == self.TD

    def __ne__(self, other) -> bool:
        if other and isinstance(other, ObjectProxy):
            return other.id != self.id or other.TD != self.TD
        return True

    def __hash__(self) -> int:
        return hash(self.id)

    # @abstractmethod
    # def is_supported_interaction(self, td, name):
    #     """Returns True if the any of the Forms for the Interaction
    #     with the given name is supported in this Protocol Binding client."""
    #     raise NotImplementedError()

    def invoke_action(self, name: str, *args, **kwargs) -> Any:
        """
        invoke an action specified by name on the server with positional/keyword arguments

        Parameters
        ----------
        name: str
            name of the action
        oneway: bool, optional, default False
            only send an instruction to invoke the action but do not fetch the reply.
            only accepted as keyword argument.
        noblock: bool, optional, default False
            schedule an action invokation but collect the reply later using a reply id.
            only accepted as keyword argument.
        *args: Any
            arguments for the action
        **kwargs: dict[str, Any]
            keyword arguments for the action

        Returns
        -------
        Any
            return value of the action call or a message id if `noblock` is True

        Raises
        ------
        AttributeError
            if action with specified name not found in the Thing Description
        Exception
            server raised exception are propagated
        """
        action = getattr(self, name, None)  # type: ConsumedThingAction
        if not isinstance(action, ConsumedThingAction):
            raise AttributeError(f"No action named {name} in Thing {self.td['id']}")
        oneway = kwargs.pop("oneway", False)
        noblock = kwargs.pop("noblock", False)
        if noblock:
            return action.noblock(*args, **kwargs)
        elif oneway:
            action.oneway(*args, **kwargs)
        else:
            return action(*args, **kwargs)

    async def async_invoke_action(self, name: str, *args, **kwargs) -> Any:
        """
        async(io) call an action specified by name on the server with positional/keyword
        arguments. `noblock` and `oneway` are not supported for async calls.

        Parameters
        ----------
        name: str
            name of the action
        *args: Any
            arguments for the action
        **kwargs: dict[str, Any]
            keyword arguments for the action

        Returns
        -------
        Any
            return value of the action call

        Raises
        ------
        AttributeError
            if action with specified name not found in the Thing Description
        Exception
            server raised exception are propagated
        """
        action = getattr(self, name, None)  # type: ConsumedThingAction
        if not isinstance(action, ConsumedThingAction):
            raise AttributeError(f"No remote action named {name}")
        return await action.async_call(*args, **kwargs)

    def read_property(self, name: str, noblock: bool = False) -> Any:
        """
        read property specified by name on server.

        Parameters
        ----------
        name: str
            name of the property
        noblock: bool, default False
            request the property but collect the reply/value later using a reply id

        Raises
        ------
        AttributeError
            if no property with specified name found in the Thing Description
        Exception
            server raised exception are propagated
        """
        prop = self.__dict__.get(name, None)  # type: ConsumedThingProperty
        if not isinstance(prop, ConsumedThingProperty):
            raise AttributeError(f"No property named {name}")
        if noblock:
            return prop.noblock_get()
        else:
            return prop.get()

    def write_property(self, name: str, value: Any, oneway: bool = False, noblock: bool = False) -> None:
        """
        write property specified by name on server with given value.

        Parameters
        ----------
        name: str
            name of the property
        value: Any
            value of property to be set
        oneway: bool, default False
            only send an instruction to write the property but do not fetch the reply.
            (irrespective of whether write was successful or not)
        noblock: bool, default False
            request the write property but collect the reply later using a reply id

        Raises
        ------
        AttributeError
            if no property with specified name found in the Thing Description
        Exception
            server raised exception are propagated
        """
        prop = self.__dict__.get(name, None)  # type: ConsumedThingProperty
        if not isinstance(prop, ConsumedThingProperty):
            raise AttributeError(f"No property named {name}")
        if oneway:
            prop.oneway_set(value)
        elif noblock:
            return prop.noblock_set(value)
        else:
            prop.set(value)

    async def async_read_property(self, name: str) -> Any:
        """
        async(io) read property specified by name on server.
        `noblock` and `oneway` are not supported for async calls.

        Parameters
        ----------
        name: Any
            name of the property to fetch

        Raises
        ------
        AttributeError
            if no property with specified name found in the Thing Description
        Exception
            server raised exception are propagated
        """
        prop = self.__dict__.get(name, None)  # type: ConsumedThingProperty
        if not isinstance(prop, ConsumedThingProperty):
            raise AttributeError(f"No property named {name}")
        return await prop.async_get()

    async def async_write_property(self, name: str, value: Any) -> None:
        """
        async(io) write property specified by name on server with specified value.
        `noblock` and `oneway` are not supported for async calls.

        Parameters
        ----------
        name: str
            name of the property
        value: Any
            value of the property to be written

        Raises
        ------
        AttributeError
            if no property with specified name found in the Thing Description
        Exception
            server raised exception are propagated
        """
        prop = self.__dict__.get(name, None)  # type: ConsumedThingProperty
        if not isinstance(prop, ConsumedThingProperty):
            raise AttributeError(f"No property named {name}")
        await prop.async_set(value)

    def read_multiple_properties(self, names: list[str], noblock: bool = False) -> dict[str, Any]:
        """
        read properties specified by list of names.

        Parameters
        ----------
        names: List[str]
            names of properties to be fetched
        noblock: bool, default False
            request the fetch but collect the reply later using a reply id

        Returns
        -------
        dict[str, Any]
            dictionary with names as keys and values corresponding to those keys

        Raises
        ------
        AttributeError
            if no property with specified name found in the Thing Description
        Exception
            server raised exception are propagated
        """
        method = getattr(self, "_get_properties", None)  # type: ConsumedThingAction
        if not method:
            raise RuntimeError("Client did not load server resources correctly. Report issue at github.")
        if noblock:
            return method.noblock(names=names)
        else:
            return method(names=names)

    def write_multiple_properties(
        self,
        oneway: bool = False,
        noblock: bool = False,
        **properties: dict[str, Any],
    ) -> None:
        """
        write properties whose name is specified as keyword arguments

        Parameters
        ----------
        oneway: bool, default False
            only send an instruction to write the property but do not fetch the reply.
            (irrespective of whether write was successful or not)
        noblock: bool, default False
            request the write property but collect the reply later using a reply id
        **properties: Dict[str, Any]
            name and value of properties to be written

        Raises
        ------
        AttributeError
            if no property with specified name found in the Thing Description
        Exception
            server raised exception are propagated
        """
        if len(properties) == 0:
            raise ValueError("no properties given to set_properties")
        method = getattr(self, "_set_properties", None)  # type: ConsumedThingAction
        if not method:
            raise RuntimeError("Client did not load server resources correctly. Report issue at github.")
        if oneway:
            method.oneway(**properties)
        elif noblock:
            return method.noblock(**properties)
        else:
            return method(**properties)

    async def async_read_multiple_properties(self, names: list[str]) -> dict[str, Any]:
        """
        async(io) read properties specified by list of names. `noblock` reads are not supported for asyncio.

        Parameters
        ----------
        names: List[str]
            names of properties to be fetched

        Returns
        -------
        dict[str, Any]
            dictionary with property names as keys and values corresponding to those keys
        """
        # TODO, actually noblock could be fine for async calls too
        method = getattr(self, "_get_properties", None)  # type: ConsumedThingAction
        if not method:
            raise RuntimeError("Client did not load server resources correctly. Report issue at github.")
        return await method.async_call(names=names)

    async def async_write_multiple_properties(self, **properties: dict[str, Any]) -> None:
        """
        async(io) write properties whose name is specified by keys of a dictionary

        Parameters
        ----------
        properties: dict[str, Any]
            name and value of properties to be written

        Raises
        ------
        AttributeError
            if no property with specified name found in the Thing Description
        Exception
            server raised exception are propagated
        """
        if len(properties) == 0:
            raise ValueError("no properties given to set_properties")
        method = getattr(self, "_set_properties", None)  # type: ConsumedThingAction
        if not method:
            raise RuntimeError("Client did not load server resources correctly. Report issue at github.")
        await method.async_call(**properties)

    def observe_property(
        self,
        name: str,
        callbacks: list[Callable] | Callable,
        asynch: bool = False,
        concurrent: bool = False,
        deserialize: bool = True,
    ) -> None:
        """
        observe a property specified by name for change events.

        Parameters
        ----------
        name: str
            name of the property
        callbacks: Callable | List[Callable]
            one or more callbacks that will be executed when the property changes
        asynch: bool
            whether the event should be listened as an asyncio task
        concurrent: bool
            - when asynch is `False`, whether to thread each of the callbacks otherwise the callbacks will be executed serially
            - when asynch is `True`, whether to create a new task for each callback otherwise the callbacks will be awaited serially
        deserialize: bool
            whether to deserialize the event data before passing it to the callbacks

        Raises
        ------
        AttributeError
            if no property with specified name found in the Thing Description or if the property is not observable
        """
        event = getattr(self, f"{name}_change_event", None)  # type: ConsumedThingEvent
        if not isinstance(event, ConsumedThingEvent):
            raise AttributeError(f"No events for property {name} or property not found")
        self.subscribe_event(
            name=f"{name}_change_event",
            callbacks=callbacks,
            asynch=asynch,
            concurrent=concurrent,
            deserialize=deserialize,
        )

    def unobserve_property(self, name: str) -> None:
        """
        Unsubscribe to property specified by name.

        Parameters
        ----------
        name: str
            name of the property

        Raises
        ------
        AttributeError
            if no property with specified name found in the Thing Description or if the property is not observable
        """
        event = getattr(self, f"{name}_change_event", None)  # type: ConsumedThingEvent
        if not isinstance(event, ConsumedThingEvent):
            raise AttributeError(f"No events for property {name} or property not found")
        event.unsubscribe()

    def subscribe_event(
        self,
        name: str,
        callbacks: list[Callable] | Callable,
        asynch: bool = False,
        concurrent: bool = False,
        deserialize: bool = True,
        # create_new_connection: bool = False,
    ) -> None:
        """
        Subscribe to event specified by name. Events are listened in separate threads and supplied callbacks are
        are also called in those threads.

        Parameters
        ----------
        name: str
            name of the event, either the object name used in the server or the name specified in the name argument of
            the Event object
        callbacks: Callable | List[Callable]
            one or more callbacks that will be executed when the event is received
        asynch: bool
            whether the event should be listened as an asyncio task
        concurrent: bool
            - when asynch is `False`, whether to thread the callbacks otherwise the callbacks will be executed serially
            - when asynch is `True`, whether to create a new task for each callback otherwise the callbacks will be awaited serially
        deserialize: bool
            whether to deserialize the event data before passing it to the callbacks

        Raises
        ------
        AttributeError
            if no event with specified name is found
        """
        event = getattr(self, name, None)  # type: ConsumedThingEvent
        if not isinstance(event, ConsumedThingEvent):
            raise AttributeError(f"No event named {name}")
        # TODO: fix the logic below to reuse connections when possible
        # if not create_new_connection:
        # see logic in tag v0.3.2
        event.subscribe(
            callbacks,
            asynch=asynch,
            concurrent=concurrent,
            deserialize=deserialize,
            # create_new_connection=create_new_connection,
        )

    def unsubscribe_event(self, name: str) -> None:
        """
        Unsubscribe to event specified by name.

        Parameters
        ----------
        name: str
            name of the event

        Raises
        ------
        AttributeError
            if no event with specified name is found
        """
        event = getattr(self, name, None)  # type: ConsumedThingEvent
        if not isinstance(event, ConsumedThingEvent):
            raise AttributeError(f"No event named {name}")
        event.unsubscribe()

    def read_reply(self, message_id: str, timeout: float | None = 5.0) -> Any:
        """
        read reply of no block calls of an action or a property read/write.

        Parameters
        ----------
        message_id: str
            id returned by the no block call
        timeout: float, optional, default 5.0
            time to wait for a reply before raising TimeoutError. None waits indefinitely.
        """
        obj = self._noblock_messages.get(message_id, None)
        if not obj:
            raise ValueError("given message id not a one way call or invalid.")
        return obj.read_reply(message_id=message_id, timeout=timeout)

    @property
    def properties(self) -> list[ConsumedThingProperty]:
        """list of properties that were consumed from the Thing Description"""
        return [prop for prop in self.__dict__.values() if isinstance(prop, ConsumedThingProperty)]

    @property
    def actions(self) -> list[ConsumedThingAction]:
        """list of actions that were consumed from the Thing Description"""
        return [action for action in self.__dict__.values() if isinstance(action, ConsumedThingAction)]

    @property
    def events(self) -> list[ConsumedThingEvent]:
        """list of events that were consumed from the Thing Description"""
        return [event for event in self.__dict__.values() if isinstance(event, ConsumedThingEvent)]

    @property
    def thing_id(self) -> str:
        """thing ID this client is connected to"""
        return self.td.get("id", None)

    @property
    def TD(self) -> dict[str, Any]:
        """Thing Description of the consuimed thing"""
        return self.td


__all__ = [ObjectProxy.__name__]



import copy

from typing import Any

import structlog

from ...constants import JSONSerializable, Operations
from ...core.zmq.message import ERROR, INVALID_MESSAGE, TIMEOUT
from ...serializers import Serializers
from ...serializers.payloads import SerializableData
from ...td import (
    ActionAffordance,
    EventAffordance,
    InteractionAffordance,
    PropertyAffordance,
)
from ...td.forms import Form
from ..repository import BrokerThing  # noqa: F401
from ..security import (
    APIKeySecurity,
    Argon2BasicSecurity,
    BcryptBasicSecurity,
    OIDCSecurity,
)


__error_message_types__ = [TIMEOUT, ERROR, INVALID_MESSAGE]


class ThingDescriptionService:
    """Service layer to generate HTTP TD"""

    def __init__(
        self,
        resource: InteractionAffordance,
        logger: structlog.stdlib.BoundLogger,
        config: Any,
        server: Any,
    ) -> None:
        from . import HTTPServer  # noqa: F401
        from .config import RuntimeConfig  # noqa: F401

        self.resource = resource  # type: InteractionAffordance
        self.config = config  # type: RuntimeConfig
        self.logger = logger.bind(layer="service", impl=self.__class__.__name__)
        self.thing = self.config.thing_repository[self.resource.thing_id]  # type: BrokerThing
        self.server = server  # type: HTTPServer

    async def generate(
        self,
        ignore_errors: bool = False,
        skip_names: list[str] = [],
        use_localhost: bool = False,
        authority: str = None,
    ) -> dict[str, JSONSerializable]:
        """
        generate the HTTP Thing Description

        Parameters
        ----------
        ignore_errors: bool, default `False`
            if `True`, errors while generating metadata for an affordances is ignored
        skip_names: list[str], default `[]`
            list of affordance names to skip while generating the TD
        use_localhost: bool, default `False`
            if `True`, localhost is used in the TD URLs instead of the server's hostname.
        authority: str, optional
            custom authority (protocol + host + port) to be used in the TD URLs. If None, the machine's hostname is used.
        """
        ZMQ_TD = await self.get_ZMQ_TD(ignore_errors=ignore_errors, skip_names=skip_names)
        TD = copy.deepcopy(ZMQ_TD)

        self.add_properties(TD, ZMQ_TD, authority=authority, ignore_errors=ignore_errors, use_localhost=use_localhost)
        self.add_actions(TD, ZMQ_TD, authority=authority, ignore_errors=ignore_errors, use_localhost=use_localhost)
        self.add_events(TD, ZMQ_TD, authority=authority, ignore_errors=ignore_errors, use_localhost=use_localhost)
        self.add_top_level_forms(TD, authority=authority, use_localhost=use_localhost)
        self.add_security_definitions(TD)
        self.add_links(TD)

        return TD

    def add_properties(
        self,
        TD: dict[str, JSONSerializable],
        ZMQ_TD: dict[str, JSONSerializable],
        authority: str,
        ignore_errors: bool,
        use_localhost: bool,
    ) -> None:
        """
        add properties to the TD with forms

        Parameters
        ----------
        TD: dict[str, JSONSerializable]
            The Thing Description to which properties are to be added
        ZMQ_TD: dict[str, JSONSerializable]
            The ZMQ Thing Description from which properties are to be read
        authority: str
            authority (protocol + host + port) to be used in the TD URLs
        ignore_errors: bool
            if `True`, errors while generating metadata for an affordances is ignored
        use_localhost: bool
            if `True`, localhost is used in the TD URLs instead of the server's hostname
        """
        from .config import HandlerMetadata

        for name in ZMQ_TD.get("properties", []):
            affordance = PropertyAffordance.from_TD(name, ZMQ_TD)
            TD["properties"][name]["forms"] = []
            try:
                href = self.server.router.get_href_for_affordance(
                    affordance,
                    authority=authority,
                    use_localhost=use_localhost,
                )
                http_methods = (
                    self.server.router.get_injected_dependencies(affordance)
                    .get("metadata", HandlerMetadata())
                    .http_methods
                )  # type: tuple[str]
            except ValueError as ex:
                if ignore_errors:
                    self.logger.warning(f"could not get HTTP methods for property {name}, skipping...")
                    continue
                raise ex from None
            for http_method in http_methods:
                if http_method.upper() == "DELETE":
                    # currently not in spec although we support it
                    continue
                if affordance.readOnly and http_method.upper() != "GET":
                    break
                op = Operations.readproperty if http_method.upper() == "GET" else Operations.writeproperty
                form = affordance.retrieve_form(op)
                if not form:
                    form = Form()
                    form.op = op
                    form.contentType = Serializers.for_object(TD["id"], TD["title"], affordance.name).content_type
                form.href = href
                form.htv_methodName = http_method
                TD["properties"][name]["forms"].append(form.json())
            if affordance.observable:
                form = affordance.retrieve_form(Operations.observeproperty)
                if not form:
                    form = Form()
                    form.contentType = Serializers.for_object(TD["id"], TD["title"], affordance.name).content_type
                    form.op = Operations.observeproperty
                form.href = f"{href}/change-event"
                form.htv_methodName = "GET"
                form.subprotocol = "sse"
                TD["properties"][name]["forms"].append(form.json())

    def add_actions(
        self,
        TD: dict[str, JSONSerializable],
        ZMQ_TD: dict[str, JSONSerializable],
        authority: str,
        ignore_errors: bool,
        use_localhost: bool,
    ) -> None:
        """
        add actions to the TD with forms

        Parameters
        ----------
        TD: dict[str, JSONSerializable]
            The Thing Description to which actions are to be added
        ZMQ_TD: dict[str, JSONSerializable]
            The ZMQ Thing Description from which actions are to be read
        authority: str
            authority (protocol + host + port) to be used in the TD URLs
        ignore_errors: bool
            if `True`, errors while generating metadata for an affordances is ignored
        use_localhost: bool
            if `True`, localhost is used in the TD URLs instead of the server's hostname
        """
        from .config import HandlerMetadata

        for name in ZMQ_TD.get("actions", []):
            affordance = ActionAffordance.from_TD(name, ZMQ_TD)
            TD["actions"][name]["forms"] = []
            try:
                href = self.server.router.get_href_for_affordance(
                    affordance,
                    authority=authority,
                    use_localhost=use_localhost,
                )
                http_methods = (
                    self.server.router.get_injected_dependencies(affordance)
                    .get("metadata", HandlerMetadata())
                    .http_methods
                )  # type: tuple[str]
            except ValueError as ex:
                if ignore_errors:
                    self.logger.warning(f"could not get HTTP methods for action {name}, skipping...")
                    continue
                raise ex from None
            for http_method in http_methods:
                form = affordance.retrieve_form(Operations.invokeaction)
                if not form:
                    form = Form()
                    form.op = Operations.invokeaction
                    form.contentType = Serializers.for_object(TD["id"], TD["title"], affordance.name).content_type
                form.href = href
                form.htv_methodName = http_method
                TD["actions"][name]["forms"].append(form.json())

    def add_events(
        self,
        TD: dict[str, JSONSerializable],
        ZMQ_TD: dict[str, JSONSerializable],
        authority: str,
        ignore_errors: bool,
        use_localhost: bool,
    ) -> None:
        """
        add events to the TD with forms

        Parameters
        ----------
        TD: dict[str, JSONSerializable]
            The Thing Description to which events are to be added
        ZMQ_TD: dict[str, JSONSerializable]
            The ZMQ Thing Description from which events are to be read
        authority: str
            authority (protocol + host + port) to be used in the TD URLs
        ignore_errors: bool
            if `True`, errors while generating metadata for an affordances is ignored
        use_localhost: bool
            if `True`, localhost is used in the TD URLs instead of the server's hostname
        """
        from .config import HandlerMetadata

        for name in ZMQ_TD.get("events", []):
            affordance = EventAffordance.from_TD(name, ZMQ_TD)
            TD["events"][name]["forms"] = []
            try:
                href = self.server.router.get_href_for_affordance(
                    affordance,
                    authority=authority,
                    use_localhost=use_localhost,
                )
                http_methods = (
                    self.server.router.get_injected_dependencies(affordance)
                    .get("metadata", HandlerMetadata(http_methods=["GET"]))
                    .http_methods
                )  # type: tuple[str]
            except ValueError as ex:
                if ignore_errors:
                    self.logger.warning(f"could not get HTTP methods for event {name}, skipping...")
                    continue
                raise ex from None
            for http_method in http_methods:
                form = affordance.retrieve_form(Operations.subscribeevent)
                if not form:
                    form = Form()
                    form.op = Operations.subscribeevent
                    form.contentType = Serializers.for_object(TD["id"], TD["title"], affordance.name).content_type
                form.href = href
                form.htv_methodName = http_method
                form.subprotocol = "sse"
                TD["events"][name]["forms"].append(form.json())

    def add_top_level_forms(
        self,
        TD: dict[str, JSONSerializable],
        authority: str,
        use_localhost: bool,
    ) -> None:
        """adds top level forms for reading and writing multiple properties"""

        properties_end_point = f"{self.server.router.get_basepath(authority, use_localhost)}/{TD['id']}/properties"

        if TD.get("forms", None) is None:
            TD["forms"] = []

        readallproperties = Form()
        readallproperties.href = properties_end_point
        readallproperties.op = "readallproperties"
        readallproperties.htv_methodName = "GET"
        readallproperties.contentType = "application/json"
        TD["forms"].append(readallproperties.json())

        writeallproperties = Form()
        writeallproperties.href = properties_end_point
        writeallproperties.op = "writeallproperties"
        writeallproperties.htv_methodName = "PUT"
        writeallproperties.contentType = "application/json"
        TD["forms"].append(writeallproperties.json())

        readmultipleproperties = Form()
        readmultipleproperties.href = properties_end_point
        readmultipleproperties.op = "readmultipleproperties"
        readmultipleproperties.htv_methodName = "GET"
        readmultipleproperties.contentType = "application/json"
        TD["forms"].append(readmultipleproperties.json())

        writemultipleproperties = Form()
        writemultipleproperties.href = properties_end_point
        writemultipleproperties.op = "writemultipleproperties"
        writemultipleproperties.htv_methodName = "PATCH"
        writemultipleproperties.contentType = "application/json"
        TD["forms"].append(writemultipleproperties.json())

    def add_security_definitions(self, TD: dict[str, JSONSerializable]) -> None:
        """adds security definitions to the TD"""
        from ...td.security_definitions import (
            APIKeySecurityScheme,
            BasicSecurityScheme,
            NoSecurityScheme,
            OIDCSecurityScheme,
        )

        TD["securityDefinitions"] = dict()

        if not self.server.config.security_schemes:
            nosec = NoSecurityScheme()
            nosec.build()
            TD["security"] = ["nosec"]
            TD["securityDefinitions"]["nosec"] = nosec.json()
            return

        TD["security"] = []
        for scheme in self.server.config.security_schemes:
            if isinstance(scheme, (BcryptBasicSecurity, Argon2BasicSecurity)):
                sec = BasicSecurityScheme()
                sec.build()
                TD["securityDefinitions"][scheme.name] = sec.json()
                TD["security"].append(scheme.name)
            if isinstance(scheme, APIKeySecurity):
                sec = APIKeySecurityScheme()
                TD["securityDefinitions"][scheme.name] = sec.json()
                TD["security"].append(scheme.name)
            if isinstance(scheme, OIDCSecurity):
                oidc_sec = OIDCSecurityScheme()
                oidc_sec.token = f"{scheme.issuer}/protocol/openid-connect/token"
                TD["securityDefinitions"][scheme.name] = oidc_sec.json()
                TD["security"].append(scheme.name)

    def add_links(self, TD: dict[str, JSONSerializable]) -> None:
        """adds custom links to the TD, override this in subclass"""
        pass

    async def get_ZMQ_TD(self, ignore_errors: bool = False, skip_names: list[str] = []) -> dict[str, JSONSerializable]:
        """fetch the TM or ZMQ in process queue TD"""
        response_message = await self.thing.execute(
            objekt=self.resource.name,
            operation=Operations.invokeaction,
            payload=SerializableData(value=dict(ignore_errors=ignore_errors, skip_names=skip_names, protocol="INPROC")),
        )
        if response_message.type in __error_message_types__:
            raise RuntimeError(f"error while fetching TD from thing - got {response_message.type} response")

        payload = self.thing.get_response_payload(response_message)
        if not isinstance(payload, SerializableData):
            raise ValueError("invalid payload received from thing")

        payload = payload.deserialize()
        if not isinstance(payload, dict):
            raise ValueError("invalid payload received from thing")
        return payload



from typing import Any, Optional

from pydantic import Field

from ..constants import JSON
from .base import Schema


class ExpectedResponse(Schema):
    """
    Form property.
    schema - https://www.w3.org/TR/wot-thing-description11/#expectedresponse
    """

    contentType: str

    def __init__(self):
        super().__init__()


class AdditionalExpectedResponse(Schema):
    """
    Form field for additional responses which are different from the usual response.
    schema - https://www.w3.org/TR/wot-thing-description11/#additionalexpectedresponse
    """

    success: bool = Field(default=False)
    contentType: str = Field(default="application/json")
    response_schema: Optional[JSON] = Field(default="exception", alias="schema")

    def __init__(self):
        super().__init__()


class Form(Schema):
    """
    Form hypermedia.
    schema - https://www.w3.org/TR/wot-thing-description11/#form
    """

    href: str = None
    op: str = None
    htv_methodName: str = Field(default=None, alias="htv:methodName")
    mqv_topic: str = Field(default=None, alias="mqv:topic")
    contentType: Optional[str] = "application/json"
    additionalResponses: Optional[list[AdditionalExpectedResponse]] = None
    contentEncoding: Optional[str] = None
    security: Optional[str] = None
    scopes: Optional[str] = None
    response: Optional[ExpectedResponse] = None
    subprotocol: Optional[str] = None

    def __init__(self):
        super().__init__()

    @classmethod
    def from_TD(cls, form_json: dict[str, Any]) -> "Form":
        """
        Create a Form instance from a Thing Description JSON object.

        Parameters
        ----------
        form_json: dict[str, Any]
            The JSON representation of the form.

        Returns
        -------
        Form
            An instance of Form.
        """
        form = cls()
        for field in cls.model_fields:
            if field == "htv_methodName" and "htv:methodName" in form_json:
                setattr(form, field, form_json["htv:methodName"])
            elif field == "mqv_topic" and "mqv:topic" in form_json:
                setattr(form, field, form_json["mqv:topic"])
            elif field in form_json:
                setattr(form, field, form_json[field])
        return form

    def __str__(self) -> str:
        return f"Form(href={self.href}, op={self.op}, htv_methodName={self.htv_methodName}, contentType={self.contentType})"



"""
The following is a list of all dataclasses used to store information on the exposed
resources on the network. These classese are generally not for consumption by the package-end-user.
"""

from enum import Enum
from types import FunctionType, MethodType
from typing import Any

from pydantic import BaseModel, RootModel

from ..constants import USE_OBJECT_NAME
from ..param.parameterized import ParameterizedMetaclass
from ..param.parameters import Boolean, ClassSelector, Parameter, String, Tuple
from ..schema_validators import BaseSchemaValidator
from ..utils import issubklass


# TODO, this class will be removed in future and merged directly into the corresponding object
class RemoteResourceInfoValidator:
    """
    A validator class for saving remote access related information on a resource. Currently callables (functions,
    methods and those with__call__) and class/instance property store this information as their own attribute under
    the variable ``_execution_info_validator``. This is later split into information suitable for HTTP server, ZMQ client & ``EventLoop``.

    Attributes
    ----------
    state : str, default None
        State machine state at which a callable will be executed or attribute/property can be
        written. Does not apply to read-only attributes/properties.
    obj_name : str, default - extracted object name
        the name of the object which will be supplied to the ``ObjectProxy`` class to populate
        its own namespace. For HTTP clients, HTTP method and URL path is important and for
        object proxies clients, the obj_name is important.
    isaction : bool, default False
        True for a method or function or callable
    isproperty : bool, default False
        True for a property
    """

    state = Tuple(
        default=None,
        item_type=(Enum, str),
        allow_None=True,
        accept_list=True,
        accept_item=True,
        doc="State machine state at which a callable will be executed or attribute/property can be written.",
    )  # type: tuple[Enum | str]

    obj = ClassSelector(
        default=None,
        allow_None=True,
        class_=(
            FunctionType,
            MethodType,
            classmethod,
            Parameter,
            ParameterizedMetaclass,
        ),  # Property will need circular import so we stick to base class Parameter
        doc="the unbound object like the unbound method",
    )

    obj_name = String(
        default=USE_OBJECT_NAME,
        doc="the name of the object which will be supplied to the ``ObjectProxy`` class to populate its own namespace.",
    )  # type: str

    isaction = Boolean(default=False, doc="True for a method or function or callable")  # type: bool

    isproperty = Boolean(default=False, doc="True for a property")  # type: bool

    def __init__(self, **kwargs) -> None:
        # No full-scale checks for unknown keyword arguments as the class
        # is used by the developer, so please try to be error-proof
        for key, value in kwargs.items():
            setattr(self, key, value)


class ActionInfoValidator(RemoteResourceInfoValidator):
    """
    request_as_argument : bool, default False
        if True, http/ZMQ request object will be passed as an argument to the callable.
        The user is warned to not use this generally.
    argument_schema: JSON, default None
        JSON schema validations for arguments of a callable. Assumption is therefore arguments will be JSON complaint.
    return_value_schema: JSON, default None
        schema for return value of a callable. Assumption is therefore return value will be JSON complaint.
    create_task: bool, default True
        default for async methods/actions
    safe: bool, default True
        metadata information whether the action is safe to execute
    idempotent: bool, default False
        metadata information whether the action is idempotent
    synchronous: bool, default True
        metadata information whether the action is synchronous
    """

    request_as_argument = Boolean(
        default=False,
        doc="if True, http/RPC request object will be passed as an argument to the callable.",
    )  # type: bool

    argument_schema = Parameter(
        default=None,
        allow_None=True,
        # due to schema validation, this has to be a dict, and not a special dict like TypedDict
        doc="JSON schema validations for arguments of a callable",
    )

    return_value_schema = Parameter(
        default=None,
        allow_None=True,
        # due to schema validation, this has to be a dict, and not a special dict like TypedDict
        doc="schema for return value of a callable",
    )

    create_task = Boolean(default=True, doc="should a coroutine be tasked or run in the same loop?")  # type: bool

    iscoroutine = Boolean(
        default=False,  # not sure if isFuture or isCoroutine is correct, something to fix later
        doc="whether the callable should be awaited",
    )  # type: bool

    safe = Boolean(default=True, doc="metadata information whether the action is safe to execute")  # type: bool

    idempotent = Boolean(default=False, doc="metadata information whether the action is idempotent")  # type: bool

    synchronous = Boolean(default=True, doc="metadata information whether the action is synchronous")  # type: bool

    isparameterized = Boolean(default=False, doc="True for a parameterized function")  # type: bool

    isclassmethod = Boolean(default=False, doc="True for a classmethod")  # type: bool

    schema_validator = ClassSelector(
        default=None,
        allow_None=True,
        class_=BaseSchemaValidator,
        doc="schema validator for the callable if to be validated server side",
    )  # type: BaseSchemaValidator

    @argument_schema.getter
    def _get_argument_schema(self):
        return getattr(self, "_argument_schema", None)

    @argument_schema.setter
    def _set_argument_schema(self, value):
        value = self.action_payload_schema_validator(value)
        setattr(self, "_argument_schema", value)

    @return_value_schema.getter
    def _get_return_value_schema(self):
        return getattr(self, "_return_value_schema", None)

    @return_value_schema.setter
    def _set_return_value_schema(self, value):
        value = self.action_payload_schema_validator(value)
        setattr(self, "_return_value_schema", value)

    def action_payload_schema_validator(self, value: Any) -> Any:
        if value is None or isinstance(value, dict) or issubklass(value, (BaseModel, RootModel)):
            return value
        raise TypeError("Schema must be None, a dict, or a subclass of BaseModel or RootModel")



import asyncio
import logging
import threading
import warnings

from io import StringIO

import structlog

from ..config import global_config
from ..core import Action, Event, Property, Thing
from ..core.properties import ClassSelector, Integer, TypedList
from ..core.zmq.rpc_server import ZMQ_TRANSPORTS, RPCServer
from ..param import Parameterized
from ..td.interaction_affordance import (
    ActionAffordance,
    EventAffordance,
    PropertyAffordance,
)
from ..utils import (
    cancel_pending_tasks_in_current_loop,
    forkable,
    get_current_async_loop,
    uuid_hex,
)
from .repository import (
    BrokerThing,
    consume_broker_pubsub,
    consume_broker_queue,
    thing_repository,
)


class BaseProtocolServer(Parameterized):
    """
    Base class for protocol specific servers.

    Protocol implementations follow a layered approach where each protocol server is split into their
    message handlers (controllers), services (important logic), and repository (for example, `Thing` repository allows
    execution of operations over the `Thing` class). This class (& its children) represent the protocol server itself
    and is responsible for starting and stopping the protocol, deciding which `Thing`s to serve etc.
    """

    port = Integer(default=9000, bounds=(1, 65535))
    """The protocol port"""

    logger = ClassSelector(
        class_=(logging.Logger, structlog.stdlib.BoundLoggerBase),
        default=None,
        allow_None=True,
    )  # type: logging.Logger | structlog.stdlib.BoundLogger
    """Logger instance"""

    things = TypedList(default=None, allow_None=True, item_type=Thing)  # type: list[Thing] | None
    """List of things to be served"""

    def __init__(self, **kwargs) -> None:
        super().__init__(**kwargs)
        if self.things is None:
            self.things = []
        self._disconnected_things = []  # type: list[BrokerThing]
        self.zmq_client_pool = None

    def add_thing(self, thing: Thing) -> None:
        """Adds a thing to the list of things to serve."""
        raise NotImplementedError("Not implemented for this protocol")

    def add_things(self, *things: Thing) -> None:
        """Adds multiple things to the list of things to serve."""
        for thing in things:
            self.add_thing(thing)

    def add_property(self, property: PropertyAffordance | Property) -> None:
        raise NotImplementedError("Not implemented for this protocol")

    def add_action(self, action: ActionAffordance | Action) -> None:
        raise NotImplementedError("Not implemented for this protocol")

    def add_event(self, event: EventAffordance | Event) -> None:
        raise NotImplementedError("Not implemented for this protocol")

    async def _instantiate_broker(
        self,
        server_id: str,
        thing_id: str,
        access_point: str = "INPROC",
    ) -> None:
        try:
            broker_thing = BrokerThing(server_id=server_id, id=thing_id, access_point=access_point)

            self._disconnected_things.append(broker_thing)

            client, TD = await consume_broker_queue(
                id=self._IP,
                server_id=server_id,
                thing_id=thing_id,
                access_point=access_point,
            )

            event_consumer = consume_broker_pubsub(
                id=self._IP,
                access_point=f"{client.socket_address}/event-publisher",
            )

            self._disconnected_things.remove(broker_thing)

            broker_thing.set_req_rep_client(client)
            broker_thing.set_event_consumer(event_consumer)
            broker_thing.TD = TD
            broker_thing.logger = structlog.get_logger().bind(
                layer="repository",
                impl=broker_thing.__class__.__name__,
                thing_id=thing_id,
            )

            if self.zmq_client_pool:
                self.zmq_client_pool.register(client, thing_id)
                broker_thing.req_rep_client = self.zmq_client_pool

            thing_repository[thing_id] = broker_thing

        except ConnectionError:
            self.logger.warning(
                f"could not connect to {thing_id} on server {server_id} with access_point {access_point}"
            )
        except Exception as ex:
            self.logger.error(f"could not connect to {thing_id} on server {server_id} with access_point {access_point}")
            self.logger.exception(ex)

    async def setup(self) -> None:
        # This method should not block, just create side-effects
        raise NotImplementedError("Not implemented for this protocol")

    async def start(self) -> None:
        # This method should not block, just create side-effects
        # await self.setup()  # call setup() here, this is only an example
        raise NotImplementedError("Not implemented for this protocol")

    @forkable
    def run(self, forked: bool = False, print_welcome_message: bool = True) -> None:
        """
        Run the server and serve your things

        Parameters
        ----------
        forked: bool, default False
            whether to run in a forked thread
        print_welcome_message: bool, default True
            whether to print a welcome message on startup, like the ports and access points
        """
        from . import run

        run(self, print_welcome_message=print_welcome_message)

    def stop(self):
        raise NotImplementedError("Not implemented for this protocol")


@forkable
def run(*servers: BaseProtocolServer, forked: bool = False, print_welcome_message: bool = True) -> None:
    """
    run servers and serve your things

    Parameters
    ----------
    servers: BaseProtocolServer
        one or more server instances to run
    forked: bool, default False
        whether to run in a forked thread
    print_welcome_message: bool, default True
        whether to print a welcome message on startup, like the ports and access points
    """
    from . import ZMQServer

    loop = get_current_async_loop()  # initialize an event loop if it does not exist

    things = [thing for server in servers if server.things is not None for thing in server.things]
    things = list(set(things))  # remove duplicates

    zmq_servers = [server for server in servers if isinstance(server, (ZMQServer, RPCServer))]
    rpc_server = None

    if len(zmq_servers) > 1:
        raise ValueError(
            "Only one ZMQServer or RPCServer instance to be run at a time, "
            + "please add all your things to one instance"
        )
    elif len(zmq_servers) == 1:
        rpc_server = zmq_servers[0]
    else:
        rpc_server = RPCServer(
            id=f"rpc-broker-{uuid_hex()}",
            things=things,
            context=global_config.zmq_context(),
        )

    threading.Thread(target=rpc_server.run).start()

    shutdown_event = asyncio.Event()
    run.shutdown_event = shutdown_event

    async def shutdown():
        shutdown_event = run.shutdown_event
        await shutdown_event.wait()

    loop = get_current_async_loop()
    for server in servers:
        if server == rpc_server:
            continue
        loop.create_task(server.start())

    if print_welcome_message:
        _print_welcome_message(servers)

    loop.run_until_complete(shutdown())
    rpc_server.stop()
    cancel_pending_tasks_in_current_loop()


def stop():
    """shutdown all running servers started with run()"""
    if hasattr(run, "shutdown_event"):
        run.shutdown_event.set()
        return
    warnings.warn(
        "No running servers found to shutdown or possibly no shutdown event available (cannot stop)",
        category=UserWarning,
    )


def parse_params(id: str, access_points: list[tuple[str, str | int | dict | list[str]]]) -> list[BaseProtocolServer]:
    from .http import HTTPServer
    from .mqtt import MQTTPublisher
    from .zmq import ZMQServer

    if access_points is not None and not isinstance(access_points, list):
        raise TypeError("access_points must be provided as a list of tuples.")

    servers = []

    for protocol, params in access_points:
        if protocol.upper() == "HTTP":
            if isinstance(params, int):
                params = dict(port=params)
            if not isinstance(params, dict):
                raise ValueError("HTTP server parameters must be supplied as a dict or just the port as an integer.")
            http_server = HTTPServer(**params)
            servers.append(http_server)
        elif protocol.upper() == "ZMQ":
            if isinstance(params, int):
                params = dict(access_points=[f"tcp://*:{params}"])
            elif isinstance(params, (str, ZMQ_TRANSPORTS)):
                params = dict(access_points=[params])
            elif isinstance(params, list):
                params = dict(access_points=params)
            if not isinstance(params.get("access_points", None), list):
                params["access_points"] = [params["access_points"]]
            if not any(isinstance(ap, str) and ap.upper().startswith("INPROC") for ap in params["access_points"]):
                params["access_points"].append("INPROC")

            if len(params["access_points"]) == 1 and params["access_points"][0] == "INPROC":
                server = RPCServer(id=id, **params)
            else:
                server = ZMQServer(id=id, **params)
            servers.append(server)
        elif protocol.upper() == "MQTT":
            if isinstance(params, str):
                params = dict(hostname=params)
            if not isinstance(params, dict):
                raise ValueError("MQTT parameters must be supplied as a dictionary or the broker hostname as a string.")
            mqtt_publisher = MQTTPublisher(**params)
            servers.append(mqtt_publisher)
        else:
            warnings.warn(f"Unsupported protocol: {protocol}", category=UserWarning)

    return servers


def _print_welcome_message(servers: list[BaseProtocolServer]) -> None:
    """prints a welcome message to the console/log"""
    from . import HTTPServer, MQTTPublisher

    buffer = StringIO()
    buffer.write("\n" + "=" * 60 + "\n")
    buffer.write("🚀 Server Started!\n")
    buffer.write("=" * 60 + "\n")
    for server in servers:
        if isinstance(server, HTTPServer):
            buffer.write("\n📡 HTTP:\n")
            for thing in server.things:
                td_path = "/resources/wot-td?ignore_errors=true"
                buffer.write(f"   ➜ Local:   {server.router.get_basepath(use_localhost=True)}/{thing.id}{td_path}\n")
                buffer.write(f"   ➜ Network: {server.router.get_basepath()}/{thing.id}{td_path}\n")
        elif isinstance(server, MQTTPublisher):
            buffer.write("\n📡 MQTT:\n")
            buffer.write(f" • Broker:   {server.hostname}:{server.port}\n")
            for thing in server.things:
                buffer.write(f"   ➜ Topic tree: {thing.id}/thing-description\n")
    buffer.write("\n" + "=" * 60 + "\n")
    print(buffer.getvalue())
